"""
R-GMA API for Python

Provides classes and interfaces to access the R-GMA information and
monitoring system.
"""

import xml.parsers.expat, sys, types, os, time, urlparse, urllib, string, httplib, socket, signal

__version__ = '5.0.12'

def interrupt_io(signum, frame):
	"""
	Raise an exception if a connection times out

	This is only used for pre-python 2.3
	"""
	raise socket.error(-1, "Connection timed out")

# Hardcoded five minute connection timeout
_connection_timeout = 300

 # Flag determining whether to use the SIGALRM to time out connections (bad, pre-python 2.3)
 # or a socket timeout (good, post-python 2.3)
_use_alarm = 0

if hasattr(socket, "setdefaulttimeout"):
	# As of Python 2.3, this method allows us to set timeouts for IO operations
	socket.setdefaulttimeout(_connection_timeout)
else:
	# For pre-Python 2.3, we fall back to interrupting IO operations using SIGALRM.
	# This means R-GMA calls MUST occur ONLY in the main thread!
	signal.signal(signal.SIGALRM, interrupt_io)
	_use_alarm = 1

class staticmethod:
	"""
	Pre-Python2.2 hack to allow us to declare static methods.
	"""
	def __init__(self, anycallable):
		self.__call__ = anycallable

class RGMAException(Exception):
	"""
	Generic exception thrown when an error occurs within the R-GMA system.

	Where possible, more specific subclasses of RGMAException will be thrown
	instead. An internal error that does not relate to user input, however, will
	be reported as an RGMAException.

	properties:

	  message          - String giving the error message
	  numSuccessfulOps - Number of successful operations. Normally this property
	                     will be 0 indicating that the operation was completely
			     unsuccessful. For batch operations, the operations will
			     be attempted in the order they are specified and will stop
			     on the first failure. In this case, numSuccessfulOps will
			     be the number of operations completed before the first
			     failure.
	  cause            - Optional exception object giving the root cause of the
	                     error. May or may not be an instance of RGMAException.
	"""
	
	def __init__(self, message, numSuccessfulOps=0, cause=None):
		"""
		Constructor.
		
		message - Message describing the error
		"""
		Exception.__init__(self, message)
		self.message = str(message)
		self.numSuccessfulOps = int(numSuccessfulOps)
		self.cause = cause
		
	def __str__(self):
		return "%s" % self.message

	def stackTrace(self):
		"""
		Returns a string containing the complete stack trace of the exception.
		"""
		stack = [str(self)]
		self.__addToStack(stack, self.cause)
		stack.reverse()
		return string.join(stack, "\n")
	
	def __addToStack(self, stack, exception):
		if exception is not None:
			stack.append(str(exception))
			if isinstance(exception, RGMAException):
				self.__addToStack(stack, exception.cause)
			
class RemoteException(Exception):
	"""
	Exception thrown when a connection to a remote service fails

	properties:

	  message - String giving the error message
	"""
	def __init__(self, host, message=None):
		"""
		Constructor.

		host    - host:port of the connection which failed
		message - Message describing the error
		"""
		Exception.__init__(self, message)
		self.host = str(host)
		self.message = str(message)

	def __str__(self):
		str = "Could not contact R-GMA server at %s" % self.host
		if self.message: str += " - %s" % self.message
		return str
	
class UnknownResourceException(Exception):
	"""
	Exception thrown when a Resource ID did not match any resource known to
	the Service
	
	properties:

	  resourceId - ID of the resource which was unknown 
	"""
	def __init__(self, resourceId):
		"""
		Constructor.
		
		resourceId - Resource ID of the unknown resource
		"""
		Exception.__init__(self, resourceId)
		self.resourceId = str(resourceId)

	def __str__(self):
		return "Could not connect to R-GMA resource: %s" % self.resourceId

class RGMASecurityException(RGMAException):
	"""
	Exception thrown when authentication or authorisation for an operation
	failed
	"""
	def __init__(self, message, numSuccessfulOps=0, cause=None):	
		RGMAException.__init__(self, message, numSuccessfulOps, cause)

class RGMAUserException(RGMAException):
	"""
	Exception thrown when there is an error relating to bad user input.

	Some examples of bad user input are:

	  Invalid tuple data for a table
	  Non-existant table
	  Non-existant VDB
	  Attempt to create a table which already exists
	  Invalid value for parameters (e.g. negative query timeout)
	"""
	def __init__(self, message, numSuccessfulOps=0, cause=None):	
		RGMAException.__init__(self, message, numSuccessfulOps, cause)

class RGMANoWorkingReplicasException(RGMAException):
	"""
	Exception thrown when there is no working schema or registry replicas on RGMA server side
	"""
	def __init__(self, message, numSuccessfulOps=0, cause=None):	
		RGMAException.__init__(self, message, numSuccessfulOps, cause)

class RGMABusyException(RGMAException):
	"""
	Exception thrown when a RGMA service is too overloaded to respond a client's request temporarily
	"""
	def __init__(self, message, numSuccessfulOps=0, cause=None):	
		RGMAException.__init__(self, message, numSuccessfulOps, cause)

class RGMABufferFullException(RGMAException):
	"""
	Exception thrown when there is insufficient space in
	a Producer to insert a tuple. The best remedy is to pause
	for some time (e.g. 5 minutes) and try again
	"""
	def __init__(self, message, numSuccessfulOps=0, cause=None):	
		RGMAException.__init__(self, message, numSuccessfulOps, cause)
	
class RGMAWarning:
	"""
	An RGMAWarning object may be attached to a ResultSet to warn about
	the quality of answers to a query.

	properties:

	  message - The warning message string
	"""
	
	def __init__(self, message):
		"""
		Constructor
		
		message - Warning message
		"""
		self.message = str(message)
				
	def getMessage(self):
		"""
		Returns the warning message
		"""
		return self.message

	def __str__(self):
		return "WARNING: %s" % self.message
	
class ResourceEndpoint:
	"""
	A resource endpoint consists of a URL and an optional resource ID

	properties:

	  url        - Service URL
	  resourceId - Unique identifier of the resource
	"""
	
	def __init__(self, url=None, resourceId=None, **kwargs):
		"""
		Create a service or resource endpoint
		
		Parameters:
		
		url        - The URL of the service or resource
		resourceId - The resource ID of the resource (optional)
		"""
		self.url = str(url)
		if resourceId is None:
			self.resourceId = None
		else:
			self.resourceId = int(resourceId)
		if kwargs.has_key('resultSet') and kwargs.has_key('fromUrl'):
			self.url = str(kwargs['fromUrl'])
			self.resourceId = int(kwargs['resultSet'][0][0])
			
	def getUrl(self):
		"""
		Get the URL of the ResourceEndpoint
		"""
		return self.url
		
	def getResourceId(self):
		"""
		Get the ID of the resource this ResourceEndpoint points to
		
		Returns None if the ResourceEndpoint has no resource ID.
		""" 
		return self.resourceId

	def __str__(self):
		ret = self.url
		if self.resourceId: ret += " %i" % self.resourceId
		return ret
	
class QueryType:
	"""
	Describes the basic type of a query.
	
	There are four types:
	
	Continuous - runs continuously and returns tuples as they are inserted.
	History    - returns tuples from producers' History queues
	Latest     - returns the most recent tuples for each primary key
	Static     - Returns tuples from On Demand producers only
	"""

	CONTINUOUS = 1
	LATEST = 2
	HISTORY = 4
	STATIC = 8

	Type = {"latest" : LATEST,
		"history" : HISTORY,
		"continuous" : CONTINUOUS,
	        "static" : STATIC}
	
	TypeName = {LATEST : "latest",
		    HISTORY : "history",
		    CONTINUOUS : "continuous",
		    STATIC : "static",
		    LATEST+HISTORY : "latest history",
		    CONTINUOUS+LATEST : "continuous latest",
		    CONTINUOUS+HISTORY : "continuous history",
		    CONTINUOUS+LATEST+HISTORY : "continuous latest history"}
	
class ProducerProperties:
	"""
	Properties of a Primary or Secondary producer
	
	A ProducerProperties object must be passed when creating a Primary
	or Secondary Producer

	properties:

	  storage - Storage object describing how the producer stores its tuples
	  queryTypes - Query types supported as ANDed set of QueryType values
	"""
	
	def __init__(self, storage, queryTypes=QueryType.CONTINUOUS, \
	             ignoreSlowConsumers=None):
		"""
		Create a ProducerProperties object
		
		Parameters:
		
		storage             - a Storage object describing how
		                      inserted tuples should be stored (e.g. 
		                      memory, database)
		queryTypes          - The types of queries to support. Should
		                      be a combination of values from 
		                      QueryType, e.g. QueryType.CONTINUOUS
		                      or QueryType.LATEST + QueryType.HISTORY 
		ignoreSlowConsumers - If True the producer will not block 
		                      inserts when a slow consumer is 
		                      preventing tuple deletion (DEPRECATED - DO
				      NOT USE)
		"""		
		self.storage = storage
		self.queryTypes = queryTypes | QueryType.CONTINUOUS
		if self.queryTypes not in QueryType.TypeName:
			raise RGMAException("Invalid query type for producer")
		
		if ignoreSlowConsumers != None:
			self.ignoreSlowConsumers = bool(ignoreSlowConsumers)
		else:
			self.ignoreSlowConsumers = False
			
	def getStorage(self):
		"""
		Return a StorageLocation object which is the type of storage 
		a producer described by these ProducerProperties will use
		"""
		return self.storage
		
	def isHistory(self):
		"""
		Returns True if these properties describe a Producer that
		can answer history queries
		"""
		return bool(self.queryTypes & QueryType.HISTORY)
		
	def isLatest(self):
		"""
		Returns True if these properties describe a Producer that
		can answer latest queries
		"""
		return bool(self.queryTypes & QueryType.LATEST)
			
	def isContinuous(self):
		"""
		Returns True if these properties describe a Producer that
		can answer latest queries
		"""
		return bool(self.queryTypes & QueryType.CONTINUOUS)
		
	def isIgnoreSlowConsumers(self):
		"""
		Returns True if these properties describe a Producer that will
		not block inserts to preserve memory if a slow consumer is 
		preventing deletion of old tuples.
		"""
		return self.ignoreSlowConsumers

	def __str__(self):
		str = QueryType.TypeName[self.queryTypes].upper()
		return str + " queries, %s" % self.storage
	
class StorageType:
	"""
	Type of storage used by a producer for its tuples.

	There are two basic types:

	  DATABASE - Tuples stored in a persistent database. This provides
	             resilience in the event of failure.
	  MEMORY   - Tuples stored in memory. Performance is likely to be
	             higher, particularly for streaming, however tuples may
		     be lost if the producer fails.
	"""
	
	FILE = 1
	DATABASE = 2
	MEMORY = 3

	Type = {"file" : FILE,
		"database" : DATABASE,
		"memory" : MEMORY}

	TypeName = {FILE : "file",
		    DATABASE : "database",
		    MEMORY : "memory"}

class TupleStore:
	"""
	A named database used by a Producer to store tuples.

	properties

	  storageType - StorageType defining the basic type of storage used
	  logicalName - Optional name for the storage, to allow a new producer to reconnect
	                to the same storage.
          location    - Optional JDBC URL of the storage, if applicable. This may allow
                        direct access to the storage if the user is authorised to access it.
	"""
	def __init__(self, storageType, logicalName=None, location=None):
		"""
		Create a StorageLocation object
		
		Parameters:
		
		storageType - Type of storage to use. Must be one of 
		              StorageType.FILE (DEPRECATED), StorageType.DATABASE or
			      StorageType.MEMORY.
		logicalName - Name of the storage. If specified, this can be used for
		              reconnecting to the same storage at a later date.
		location    - JDBC url of the database used for the DATABASE storage type.
		              Note that this cannot be specified when creating a producer, however
			      the database location of a previously created tuple store
			      can be found using the listTupleStores method on the Producer
			      instance.
		"""
		if storageType not in (StorageType.FILE, StorageType.DATABASE, StorageType.MEMORY):
			raise RGMAException("storageType must be of StorageType.FILE, StorageType.DATABASE or StorageType.MEMORY")
			
		self.type = storageType
		self.logicalName = logicalName
		self.location = location
		
class Storage:
	"""
	Storage describes how tuples are stored for a Producer
	
	It may describe a database, an in-memory structure, flat file, etc.
	"""
	
	def __init__(self, storageType, \
	             location=None, username=None, password=None):
		"""
		Create a StorageLocation object
		
		Parameters:
		
		storageType - Type of storage to use. Must be one of 
		              StorageType.FILE (DEPRECATED), StorageType.DATABASE or
			      StorageType.MEMORY.
		location    - location of the storage. This should be a logical name
		              and only used when reconnecting to a storage location
			      used by a previous Producer (optional).
			      If the storage type is DATABASE and a username and
			      password are specified, the location is interpreted as
			      a JDBC database URL. HOWEVER THIS BEHAVIOUR IS DEPRECATED AND
			      SHOULD NOT BE USED.
		username    - Username to use when accessing storage (optional)
		              THIS PARAMETER IS DEPRECATED
		password    - Password to use when accessing storage (optional)
		              THIS PARAMETER IS DEPRECATED
		"""
		if storageType not in (StorageType.FILE, StorageType.DATABASE, StorageType.MEMORY):
			raise RGMAException("storageType must be of StorageType.FILE, StorageType.DATABASE or StorageType.MEMORY")

		if location:
			self.location = str(location)
			self.logicalName = str(location)
		else:
			self.location = None
			self.logicalName = None
		if username:
			self.username = str(username)
		else:
			self.username = None
		if password:
			self.password = str(password)
		else:
			self.password = None
		self.type = storageType
			
	def getUsername(self):
		"""
		Returns the username used to access the storage, or None
		if none required
		"""
		return self.username
	
	def getPassword(self):
		"""
		Returns the password used to access the storage, or None
		if none required
		"""
		return self.password
		
	def getLocation(self):
		"""
		Returns the location (e.g. URL, file path) used to access the 
		storage, or None if not applicable
		"""
		return self.location

	def getLogicalName(self):
		"""
		Returns the logical name of the storage or None if not applicable
		"""
		return self.logicalName

	def getType(self):
		"""
		Returns the type of storage, one of StorageType.DATABASE,
		StorageType.MEMORY or StorageType.FILE
		"""
		return self.type

	def __str__(self):
		str = StorageType.TypeName[self.type].upper() + " storage"
		if self.logicalName:
			if self.username and self.password:
				str += "(url: %s)" % self.logicalName
			else:
				str += "(logical name: %s" % self.logicalName
		else:
			str += " (default location)"
		return str

class Units:
	"""
	Base class for a unit of time

	properties:

	  seconds - Number of seconds in each unit
	"""
	
	def __init__(self, seconds):
		"""
		Constructor
		
		Parameters:
		
		seconds - The number of seconds represented by this time unit
		"""
		self.seconds = float(seconds)

	def getSeconds(self):
		"""
		Return the number of seconds represented by this time unit
		"""
		return self.seconds

Units.SECONDS = Units(1)
Units.MINUTES = Units(60)
Units.HOURS = Units(60*60)
Units.DAYS = Units(24*60*60)

Units.UnitName = {Units.SECONDS : "seconds",
		  Units.MINUTES : "minutes",
		  Units.HOURS : "hours",
		  Units.DAYS : "days"}

Units.Unit = {"seconds" : Units.SECONDS,
	      "minutes" : Units.MINUTES,
	      "hours" : Units.HOURS,
	      "days" : Units.DAYS}

class TimeInterval:
	"""
	A specific length of time comprising of a length and units (e.g.
	1 second, 7 hours, etc.)

	Time intervals must be > 0
	
	properties:

	  value - Number of time units in the interval
	  units - Units the interval is measured in
	"""

	def __init__(self, value=1, units=Units.SECONDS, **kwargs):
		"""
		Constructor
		
		Parameters:
		
		value - Length of the time interval
		units - Units the length is measured in (optional - default seconds)
		"""
		self.value = float(value)
		self.units = units
		if kwargs.has_key('resultSet'):
			self.value = float(kwargs['resultSet'][0][0])
			self.units = Units.SECONDS

	def getValue(self):
		"""
		Returns the value of the time interval
		"""
		return self.value

	def getUnits(self):
		"""
		Returns the units the time interval is specified in
		"""
		return self.units
	
	def getValueAs(self, units):
		"""
		Get the length of the time interval in specified units
		
		Parameters:
		
		units - Units object specifying time units to return
		"""
		seconds = self.value * self.units.getSeconds()
		return seconds / units.getSeconds()
	
	def __str__(self):
		return "%g %s" % (self.value, Units.UnitName[self.units])
	
class QueryProperties:
	"""
	Describes the set of tuples to be queried.
	
	Each query has a basic type (e.g. QueryType.CONTINUOUS) and may have an
	associated TimeInterval representing the maximum age of tuples to 
	query.
	
	In the case of a continuous query the associated TimeInterval causes the
	consumer to initially return a history of all tuples available up to
	this maximum age, and to return all inserted tuples thereafter.
		
	properties:

	  queryType    - QueryType value describing the basic type of query.
	  timeInterval - Associated time interval giving the maximum age of
	                 tuples to return.
	"""

	def __init__(self, queryType, timeInterval=None):
		"""
		Constructor
		
		Parameters:
		
		queryType    - basic type of query. Must be one of:
		                 QueryType.CONTINUOUS
		                 QueryType.HISTORY
		                 QueryType.LATEST
		timeInterval - TimeInterval object specifying maximum age of 
		               tuples to include in the results of the query 
		               (optional - default no maximum)
		"""
		if queryType not in (QueryType.CONTINUOUS, QueryType.HISTORY, \
				     QueryType.LATEST, QueryType.STATIC):
		    raise RGMAException("queryType must be one of QueryType.CONTINUOUS, QueryType.HISTORY, QueryType.LATEST or QueryType.STATIC")
		    
		self.queryType = queryType
		self.timeInterval = timeInterval
	
	def hasTimeInterval(self):
		"""
		Returns True if the query has an associated time interval
		"""
		return self.timeInterval != None
			
	def getTimeInterval(self):
		"""
		Return TimeInterval specifying the maximum age of tuples to
		return as results of the query
		"""
		return self.timeInterval
	
	def getQueryType(self):
		"""
		Return the basic query type, i.e. one of:
		
		QueryType.CONTINUOUS
		QueryType.HISTORY
		QueryType.LATEST
		"""
		return self.queryType
	
	def isContinuous(self):
		"""
		Returns true if the query is continuous
		"""
		return self.queryType == QueryType.CONTINUOUS

	def isHistory(self):
		"""
		Returns true if the query is over archived historical tuples
		"""
		return self.queryType == QueryType.HISTORY

	def isLatest(self):
		"""
		Returns true if the query is over latest tuples only
		"""
		return self.queryType == QueryType.LATEST

	def isStatic(self):
		"""
		Returns true if the query is over latest tuples only
		"""
		return self.queryType == QueryType.STATIC

	def __str__(self):
		str = "%s query" % QueryType.TypeName[self.queryType]
		if self.timeInterval:
			str += " (time interval: %s)" % self.timeInterval
		return str
	
class Types:
	"""
	SQL data types

	Note that users should use these constants by name, rather than
	relying on their numeric values which may change.
	"""

	UNKNOWN = 0
	INTEGER = 1
	REAL = 2
	DOUBLE = 3
	CHAR = 4
	VARCHAR = 5
        TIMESTAMP = 6
	DATE = 7
	TIME = 8

	TypeName = {INTEGER : "INTEGER",
		    REAL : "REAL",
		    DOUBLE : "DOUBLE",
		    CHAR : "CHAR",
		    VARCHAR : "VARCHAR",
		    TIMESTAMP : "TIMESTAMP",
		    DATE : "DATE",
		    TIME : "TIME",
		    UNKNOWN : "UNKNOWN"}
	
	Type = {"INTEGER" : INTEGER,
		"REAL" : REAL,
		"DOUBLE" : DOUBLE, 
		"CHAR" : CHAR,
		"VARCHAR" : VARCHAR,
		"TIMESTAMP" : TIMESTAMP,
		"DATE" : DATE,
		"TIME" : TIME}

def TypeString(type, width=None):
	"""
       	Convert a type with optional width into a string
       	"""
       	try:
       		if width:
       			return "%s(%i)" % (Types.TypeName[type], width)
       		else:
       			return Types.TypeName[type]
       	except:
       		raise RGMAException("Unrecognized type number: %i" % type)

class ResultSet(list):
	"""
	A ResultSet is returned by a Consumer in response to a query
	
	This is a subclass of the standard Python list so it can be manipulated
	using the usual list functions.

	properties:

	  columnNames     - List of column name strings
	  columnTableName - List table names which each column comes from. May be empty or None
	  columnTypes     - List of Types giving the basic SQL type of each column
	  endOfResults    - If 'true', this ResultSet contains the last results for the query.
	  warning         - Warning string, if the results may be incomplete. May be None.
	  columnWidths    - Maximum character widths of the data in each column.
	"""

	def __init__(self, tuples, columnNames=[], columnTableNames=[], \
	             columnTypes=[], endOfResults=False, warning=None):
		"""
		Constructor

		tuples           - List of tuples
		columnNames      - List of column names
		columnTableNames - List of the names of tables which the columns belong to
		columnTypes      - List of rgma.Types object defining the column types
		warning          - An RGMAWarning object if the results may be incomplete
		"""
		
		if not columnNames: columnNames = []
		if not columnTableNames: columnTableNames = []
		if not columnTypes: columnTypes = []
		
		if len(columnNames) != 0 and len(columnTableNames) != 0:
			if len(columnNames) != len(columnTableNames):
				raise RGMAException("Number of columns is inconsistent")
		
		if len(columnTypes) != 0 and len(columnTableNames) != 0:
			if len(columnTypes) != len(columnTableNames):
				raise RGMAException("Number of columns is inconsistent")
			
		if len(columnTypes) != 0 and len(columnNames) != 0:
			if len(columnTypes) != len(columnNames):
				raise RGMAException("Number of columns is inconsistent")
		
		if len(tuples) > 0:
			numColumns = len(tuples[0])
			for tuple in tuples:
				if len(tuple) != numColumns:
					raise RGMAException("Tuples have inconsistent columns")
		
		if len(columnNames) != 0:
			if len(tuples) > 0:
				if numColumns != len(columnNames):
					raise RGMAException("Tuples and column names have inconsistent columns")

		if len(columnTypes) != 0:
			if len(tuples) > 0:
				if numColumns != len(columnTypes):
					raise RGMAException("Tuples and column types have inconsistent columns")

		if len(columnTableNames) != 0:
			if len(tuples) > 0:
				if numColumns != len(columnTableNames):
					raise RGMAException("Tuples and column table names have inconsistent columns")

		list.__init__(self, tuples)
		self.warning = warning
		self.columnNames = columnNames
		self.columnTableNames = columnTableNames
		self.columnTypes = columnTypes
		self.endOfResults = endOfResults
		
		# Finds maximum width for each column
		self.columnWidths = []
		if len(self) > 0:
			numColumns = len(self[0])
		else:
			numColumns = len(self.columnNames)
			
		for i in range(numColumns):
			self.columnWidths.append(2)
		col = 0
               	for columnName in self.columnNames:
			self.columnWidths[col] = (len(columnName)+2)
			col += 1
               	for row in self:
        		col = 0
            		for item in row:
            			string = item.__str__()
                		if self.columnWidths[col] < len(string) + 2:
                			self.columnWidths[col] = len(string) + 2
                		col += 1
				
	def __add__(self, rs):
		if isinstance(rs, ResultSet):
			warning = rs.warning
			end = rs.endOfResults
		else:
			warning = self.warning
			end = self.endOfResults
		return ResultSet(list.__add__(self, rs), self.columnNames, \
				 self.columnTableNames, self.columnTypes, \
				 end, warning)

	def __mul__(self, n):
		return ResultSet(list.__mul__(self, n), self.columnNames, \
				 self.columnTableNames, self.columnTypes, \
				 self.endOfResults, self.warning)

	def __rmul__(self, n):
		return self.__mul__(n)

	def __iadd__(self, rs):
		if isinstance(rs, ResultSet):
			if rs.warning != None:
				self.warning = rs.warning
			if rs.endOfResults:
				self.endOfResults = 1
		return list.__iadd__(self, rs)
	
	def getColumnNames(self):
		"""
		Get the names of the columns in the ResultSet, not including
		any table name prefix
		"""
		return self.columnNames
			
	def getColumnTableNames(self):
		"""
		Get the names of the tables each column is taken from
		"""
		return self.columnTableNames

	def getColumnTypes(self):
		"""
		Get the types of the data in the columns

		Returns rgma.Types objects, e.g. rgma.Types.INTEGER
		"""
		return self.columnTypes

	def getColumnWidths(self):
		"""
		Get a list of the maximum widths of the data in each column
		"""
		return self.columnWidths
		
	def getWarning(self):
		"""
		Return the RGMAWarning object associated with this ResultSet
		or None if no warning was associated
		"""
		return self.warning

	def headerToString(self, columnWidths=None):
		if not columnWidths:
			columnWidths = self.columnWidths
	        # Prints First Separator 
        	str = "\n+"
        	for i in range(len(self.columnNames)):
            		spacer = ""
            		for j in range(columnWidths[i]):
                		spacer = spacer + "-"
            		str =  str + spacer + "+"
		str += "\n"
		
        	#Prints Column Names
        	str += "| "
        	spacer = ""
        	col = 0
        	for column in self.columnNames:
            		str = str + column
            		padding = columnWidths[col] - len(column) - 1
            		for j in range(padding):
                		str = str + " "
            		str += "| "
            		col += 1
        	str += "\n"
						
		# Prints Second Separator
		str += "+"
        	for i in range(len(self.columnNames)):
            		spacer = ""
            		for j in range(columnWidths[i]):
                		spacer = spacer + "-"
            		str += spacer + "+"

		return str

	def rowsToString(self, columnWidths = None):
		if not columnWidths:
			columnWidths = self.columnWidths
         	# Prints Rows
		rowStrings = []
        	for row in self:
            		str = "| "
            		spacer = ""
            		col = 0
            		for item in row:
				if item is None:
					stritem = "NULL"
				else:
					stritem = item.__str__()
                		str += stritem
                		padding = columnWidths[col] - len(stritem) - 1
                		for k in range(padding):
                    			str = str + " "
                		str += "| "
                		col += 1
            		rowStrings.append(str)
			
		return string.join(rowStrings, "\n")

	def footerToString(self, columnWidths=None):
		if not columnWidths:
			columnWidths = self.columnWidths
      		# Prints Final Separator
		str = "+"
		for width in columnWidths:
            		spacer = ""
            		for j in range(width):
                		spacer = spacer + "-"
            		str += spacer + "+"
		return str
	
        def printHeader(self, columnWidths=None):
        	"""
        	Print the column headings for the result set
        	"""
		print self.headerToString(columnWidths)
		    	
      	def printRows(self, columnWidths=None):
      		"""
      		Print the rows in the result set formatted in a table
      		"""
		print self.rowsToString(columnWidths)
		
      	def printFooter(self, columnWidths=None):
      		"""
      		Print the bottom of the result set table
      		"""
		print self.footerToString(columnWidths)

        def printWarning(self):
        	if self.warning: print self.warning
        		
	def printTable(self):
		"""
		Print the whole result set as a table
		"""
		self.printHeader()
		self.printRows()
		self.printFooter()
		self.printWarning()

	def __str__(self):
		str = self.headerToString() + "\n" + self.rowsToString() + "\n" + self.footerToString()
		if self.warning: str += "\n%s" % self.warning
		return str

class _RGMATuple(tuple):
	"""
	Special RGMA version of a Python tuple to allow things
	like tuple['columnName'] as well as tuple[0]
	"""

	def __getitem__(self, item):
		if type(item) in types.StringTypes:
			try:
				index = self.columnNames.index(item)
				return tuple.__getitem__(self, index)
			except:
				raise RGMAException("No such column: %s" % item)
		else:
			return tuple.__getitem__(self, item)
		
class _ResponseParser:
	"""
	Parser for XML responses from the RGMA server
	"""

	EXCEPTION_TYPES = {0 : RGMAException,
			   1 : RGMASecurityException,
			   2 : RGMAUserException,
			   3 : RGMABufferFullException,
			   7 : RGMANoWorkingReplicasException,
			   9 : RGMABusyException}
			   			   
	TYPE_HANDLERS = {Types.VARCHAR : str,
			 Types.CHAR : str,
			 Types.INTEGER : int,
			 Types.REAL : float,
			 Types.DOUBLE : float,
			 Types.DATE : str,
			 Types.TIME : str,
			 Types.TIMESTAMP : str,
			 Types.UNKNOWN : str}

	TYPE_CODES = {4 : Types.INTEGER,
		      6 : Types.REAL,
		      7 : Types.REAL,
		      8 : Types.DOUBLE,
		      1 : Types.CHAR,
		      12 : Types.VARCHAR,
		      93 : Types.TIMESTAMP,
		      91 : Types.DATE,
		      92 : Types.TIME}
	
	def __init__(self, xmlData):
		self.TOPLEVEL_HANDLERS = {"edg:XMLResponse" : (self.responseStart, self.responseEnd)}

		self.OBJECT_HANDLERS = {"XMLResultSet" : (self.resultSetStart, self.resultSetEnd),
					"XMLException" : (self.exceptionStart, self.exceptionEnd)}
		
		self.RESULTSET_HANDLERS = {"row" : (self.rowStart, self.rowEnd),
					   "metadata" : (self.metadataStart, self.metadataEnd),
					   "rowMetaData" : (self.oldMetadataStart, self.oldMetadataEnd),
					   "RGMAWarning" : (self.warningStart, self.warningEnd)}

		self.ROW_HANDLERS = {"col" : (self.colStart, self.colEnd)}

		self.METADATA_HANDLERS = {"column" : (self.colMetadataStart, self.colMetadataEnd)}
		self.COL_METADATA_HANDLERS = {"table" : (self.tableMetadataStart, self.tableMetadataEnd),
					      "name" : (self.nameMetadataStart, self.nameMetadataEnd),
					      "type" : (self.typeMetadataStart, self.typeMetadataEnd)}

		self.OLD_METADATA_HANDLERS = {"colMetaData" : (self.oldColMetadataStart, self.oldColMetadataEnd)}
		
		self.EXCEPTION_HANDLERS = {"message" : (self.exceptionMessageStart, self.exceptionMessageEnd),
					   "XMLException" : (self.exceptionStart, self.exceptionEnd)}

		self.TOPLEVEL_HANDLERS = {"edg:XMLResponse" : (self.responseStart, self.responseEnd)}

		self.OBJECT_HANDLERS = {"XMLResultSet" : (self.resultSetStart, self.resultSetEnd),
					"XMLException" : (self.exceptionStart, self.exceptionEnd)}
	
		self.RESULTSET_HANDLERS = {"row" : (self.rowStart, self.rowEnd),
					   "metadata" : (self.metadataStart, self.metadataEnd),
					   "rowMetaData" : (self.oldMetadataStart, self.oldMetadataEnd),
					   "RGMAWarning" : (self.warningStart, self.warningEnd)}

		self.ROW_HANDLERS = {"col" : (self.colStart, self.colEnd)}

		self.METADATA_HANDLERS = {"column" : (self.colMetadataStart, self.colMetadataEnd)}
		self.COL_METADATA_HANDLERS = {"table" : (self.tableMetadataStart, self.tableMetadataEnd),
					      "name" : (self.nameMetadataStart, self.nameMetadataEnd),
					      "type" : (self.typeMetadataStart, self.typeMetadataEnd)}

		self.OLD_METADATA_HANDLERS = {"colMetaData" : (self.oldColMetadataStart, self.oldColMetadataEnd)}
		
		self.EXCEPTION_HANDLERS = {"message" : (self.exceptionMessageStart, self.exceptionMessageEnd),
					   "XMLException" : (self.exceptionStart, self.exceptionEnd)}

		self.WARNING_HANDLERS = {"message" : (self.warningMessageStart, self.warningMessageEnd)}

		self.handlers = self.TOPLEVEL_HANDLERS
		self.endHandlers = []
		self.exception = None
		self.exceptions = []
		self.warning = None
		self.endOfResults = False
		self.currentCol = 0
		self.chars = ""
		self.rows = []
		self.columnNames = []
		self.columnTableNames = []
		self.columnTypes = []
		self.endOfResults = False
		self.warning = None
		
		p = xml.parsers.expat.ParserCreate()
		p.StartElementHandler = self.startElement
		p.EndElementHandler = self.endElement
		p.CharacterDataHandler = self.charData
		try:
			p.Parse(xmlData)
		except Exception, e:
			raise RGMAException("Invalid response R-GMA server - check the server is running and your R-GMA configuration is correct", 0, e)
		
		if self.exception:
			raise self.exception
			
		self.resultSet = ResultSet(self.rows,
					   self.columnNames,
					   self.columnTableNames,
					   self.columnTypes,
					   self.endOfResults,
					   self.warning)

	## Generic SAX handlers
		
	def startElement(self, name, attrs):
		"""
		Called when an element start tag is found
		"""
		try:
			handlers = self.handlers[name]
			self.endHandlers.append(handlers[1])
			handlers[0](name, attrs)
		except KeyError:
			raise RGMAException("Invalid XML response from server: got element '%s', expected one of [%s]" \
					    % (name, string.join(self.handlers.keys(), ",")))
	def endElement(self, name):
		"""
		Called when an element end tag is found
		"""
		endHandler = self.endHandlers.pop()
		endHandler(name)

	def charData(self, data):
		"""
		Called when character data is found
		"""
		self.chars = self.chars + data

	## Handler implementations

	def responseStart(self, name, attrs):
		self.handlers = self.OBJECT_HANDLERS

	def responseEnd(self, name):
		self.handlers = self.TOPLEVEL_HANDLERS
		
	def resultSetStart(self, name, attrs):
		self.handlers = self.RESULTSET_HANDLERS
		self.endOfResults = attrs.get("endOfResults", "false").lower() == "true"
		self.rows = []

	def resultSetEnd(self, name):
		self.handlers = self.OBJECT_HANDLERS
	
	def exceptionStart(self, name, attrs):
		type = attrs.get('type', None)
		if type.rfind('RGMAException') != -1:
			errorNumber = int(attrs.get('errorNumber', 0))
			numSuccessfulOps = int(attrs.get('numSuccessfulOps', 0))
			exception = self.EXCEPTION_TYPES.get(errorNumber, RGMAException)('', numSuccessfulOps, None)
		elif type.rfind('UnknownResourceException') != -1:
			exception = UnknownResourceException('')
		else:
			exception = RGMAException('', 0, None)

		if self.exception is not None:
			self.exception.cause = exception
			self.exceptions.append(self.exception)
		self.exception = exception
		self.handlers = self.EXCEPTION_HANDLERS

	def exceptionEnd(self, name):
		self.handlers = self.OBJECT_HANDLERS
		if len(self.exceptions) > 0:
			self.exception = self.exceptions.pop()
			
	def rowStart(self, name, attrs):
		self.handlers = self.ROW_HANDLERS
		self.row = []

	def rowEnd(self, name):
		self.handlers = self.RESULTSET_HANDLERS
		tuple = _RGMATuple(self.row)
		tuple.columnNames = self.columnNames
		self.rows.append(tuple)
	
	def colStart(self, name, attrs):
		self.chars = ""
		if attrs.get("isNull", "false").lower() == "true":
			self.nullColumn = True
		else:
			self.nullColumn = False

	def colEnd(self, name):
		if len(self.columnTypes) == 0:
			type = Types.VARCHAR
		else:
			type = self.columnTypes[self.currentCol]				
			self.currentCol += 1
			if self.currentCol == len(self.columnTypes):
				self.currentCol = 0
				
		if self.nullColumn:
			self.row.append(None)
		else:
			self.row.append(self.TYPE_HANDLERS[type](self.chars))

	def metadataStart(self, name, attrs):
		self.handlers = self.METADATA_HANDLERS
		self.columnNames = []
		self.columnTypes = []
		self.columnTableNames = []

	def metadataEnd(self, name):
		self.handlers = self.RESULTSET_HANDLERS

	def colMetadataStart(self, name, attrs):
		self.handlers = self.COL_METADATA_HANDLERS

	def colMetadataEnd(self, name):
		self.handlers = self.METADATA_HANDLERS

	def tableMetadataStart(self, name, attrs):
		self.chars = ""

	def tableMetadataEnd(self, name):
		self.columnTableNames.append(self.chars)

	def nameMetadataStart(self, name, attrs):
		self.chars = ""

	def nameMetadataEnd(self, name):
		self.columnNames.append(self.chars)

	def typeMetadataStart(self, name, attrs):
		self.chars = ""

	def typeMetadataEnd(self, name):
		try:
			self.columnTypes.append(self.TYPE_CODES[int(self.chars)])
		except KeyError:
			self.columnTypes.append(Types.UNKNOWN)
	
	def oldMetadataStart(self, name, attrs):
		self.handlers = self.OLD_METADATA_HANDLERS
		self.columnNames = []
		self.columnTypes = []
		self.columnTableNames = []
		
	def oldMetadataEnd(self, name):
		self.handlers = self.RESULTSET_HANDLERS
			
	def oldColMetadataStart(self, name, attrs):
		self.chars = ""
		
	def oldColMetadataEnd(self, name):
		self.columnNames.append(self.chars)
		self.columnTableNames.append("")
		self.columnTypes.append(Types.VARCHAR)

	def warningStart(self, name, attrs):
		self.chars = ""
		self.handlers = self.WARNING_HANDLERS

	def warningEnd(self, name):
		self.handlers = self.RESULTSET_HANDLERS
		
	def warningMessageStart(self, name, attrs):
		self.chars = ""

	def warningMessageEnd(self, name):
		self.warning = RGMAWarning(self.chars)
		
	def exceptionMessageStart(self, name, attrs):
		self.chars = ""

	def exceptionMessageEnd(self, name):
		self.exception.message = self.chars
					
class Properties:
	"""
	Provides access to the name/value properties loaded from a file
	
	Typically used to access the RGMA global configuration file
	$RGMA_HOME/etc/rgma/rgma.conf
	"""

	def __init__(self, propsFile=None):
		"""
		Constructor
		
		Loads properties from a name=value format props file
		
		Parameters:
		
		propsFile - The file to load properties from. The usual usage is
		            to leave this unspecified in which case it defaults
		            to $RGMA_HOME/etc/rgma/rgma.conf (raising an 
		            exception if the environment variable RGMA_HOME is
		            unset).
		"""
		if not propsFile:
			try:
				rgmaHome = os.environ['RGMA_HOME']
			except KeyError:
				try:
					rgmaHome = os.environ['GLITE_LOCATION']
				except KeyError:
					raise RGMAException("Environment variable RGMA_HOME or GLITE_LOCATION is not set")
		
			propsFile = os.path.join(rgmaHome, "etc/rgma/rgma.conf")

		self.properties = self.getProperties(propsFile)
		
	def getProperty(self, property):
		"""
		Get a property from the RGMA configuration file

		Parameters:
		
		property - Name of the property to look up
		
		Returns:
		
		A tuple containing all values of the property or an empty tuple if it is not set
		"""
		try:
			return self.properties[property]
		except KeyError:
			return ()
		
	def getProperties(self, filename):
		"""
		Load properties from a simple properties file
		
		The file must be of the form property=value. Whitespace is
		discounted and lines beginning with # are ignored

		Returns a dictionary of key/value pairs
		"""
	        props = {}
		file = open(filename)
		lines = file.readlines()
		file.close()
		
		for line in lines:
			comment_index = line.find("#")
			actual = line[:comment_index].strip()
			equals_index = actual.find("=")
			
			if len(actual) > 0 and equals_index >= 0:
				property = actual[:equals_index].strip()
				value = actual[equals_index+1:].strip()

                                if len(property) > 0 and len(value) > 0:
					if props.has_key(property):
						props[property] = props[property] + (value,)
					else:
						props[property] = (value,)

		return props

	def __str__(self):
		str = ''
		for property in self.properties:
			for value in self.properties[property]:
				str += "%s=%s\n" % (property, value)
		return str
	
def _getSecurityInfo():
	"""
	Get the certificate and key files to be used for secure connections.

	Returns cert_file, key_file
	"""
	if os.environ.has_key('X509_USER_PROXY'):
		proxy = os.environ['X509_USER_PROXY']
		if len(proxy) > 0:
			if not os.path.exists(proxy):
				raise RGMASecurityException("X509_USER_PROXY set to file %s which does not exist" % proxy)
			if not os.access(proxy, os.R_OK):
				raise RGMASecurityException("X509_USER_PROXY set to file %s which is not readable" % proxy)
			return (proxy, proxy)
	if os.environ.has_key('TRUSTFILE'):
		trustfile = os.environ['TRUSTFILE']
		if len(trustfile) > 0:
			try:
				props = Properties(trustfile)
			except Exception, e:
				raise RGMASecurityException("Couldn't open TRUSTFILE: %s" % trustfile, 0, e)
			cert_files = props.getProperty("sslCertFile")
			key_files = props.getProperty("sslKey")
			proxy_files = props.getProperty("gridProxyFile")
			if len(cert_files) > 0 and len(key_files) > 0:
				if not os.path.exists(cert_files[0]):
					raise RGMASecurityException("sslCertFile in TRUSTFILE (%s) set to file %s which does not exist" % \
								    (trustfile, cert_files[0]))
				if not os.path.exists(key_files[0]):
					raise RGMASecurityException("sslKey in TRUSTFILE (%s) set to file %s which does not exist" % \
								    (trustfile, key_files[0]))
				return (cert_files[0], key_files[0])
			elif len(proxy_files) > 0:
				if not os.path.exists(proxy_files[0]):
					raise RGMASecurityException("gridProxyFile in TRUSTFILE (%s) set to file %s which does not exist" % \
								    (trustfile, proxy_files[0]))
				return (proxy_files[0], proxy_files[0])
			else:
				raise RGMASecurityException("Either sslCertFile and sslKey or gridProxyFile must be set in TRUSTFILE (%s)" % \
                                                    trustfile)                          

	raise RGMASecurityException("Neither TRUSTFILE nor X509_USER_PROXY is set")

def _urlencode(params):
	paramsList = []
	for key, value in params.items():
		if type(value) in (types.ListType, types.TupleType):
			for item in value:
				paramsList.append("%s=%s" % (urllib.quote_plus(key), urllib.quote_plus(item)))
		else:
			paramsList.append("%s=%s" % (urllib.quote_plus(key), urllib.quote_plus(str(value))))
			
	return string.join(paramsList, "&")

class Service:
	"""
	Base class for a connection to an R-GMA service.

	A service in itself has no state. Stateful resources are
	modelled by adding a connection ID to service method calls.

	properties:

	  url - The URL of the service endpoint as a string
	"""

	def __init__(self, serviceName, url=None):
		"""
		Constructor

		Parameters:
		  serviceName - Service name in rgma.conf (e.g. "Consumer").
		  url         - Service url (looked up in rgma.conf if omitted)
		"""

		self.conn = None
		if url == None:
			urls = Properties().getProperty(serviceName)
			if len(urls) == 0:
				raise RGMAException("No entry for " + serviceName + " in $RGMA_HOME/etc/rgma/rgma.conf")
			url = urls[0] # use the first entry only

		try:
			parsedUrl = urlparse.urlparse(url)
		except Exception, e:
			raise RGMAException("Invalid service URL: %s" % url, 0, e)

		self.url = url
		self.secure = parsedUrl[0] == 'https'
		self.host = parsedUrl[1]
		self.path = parsedUrl[2]

	def __del__(self):
		self._clear_connection()

	def _clear_connection(self):
		if hasattr(self, 'conn'):
			if self.conn != None:
				try:
					self.conn.close()
				except AssertionError, e:
					print "AssertionError when closing connection It looks like this is the buggy version of Python - please consider using a different version.It has been tested with Python 2.4.1"
				self.conn = None
	def _connect(self):
		try:
			try:
				if _use_alarm:
				
					signal.alarm(_connection_timeout)
				if self.conn == None:
					if self.secure:
						cert, key = _getSecurityInfo()
						self.conn = httplib.HTTPSConnection(self.host, key_file=key, cert_file=cert)
					else:
						self.conn = httplib.HTTPConnection(self.host)
			finally:
				if _use_alarm:
					
					signal.alarm(0) # Clear the previously set alarm
		except IOError, e:
			self._clear_connection()
			raise RemoteException(self.host, e.strerror)
		except httplib.HTTPException, e:
			self._clear_connection()
			raise RemoteException(self.host, "HTTP connection failed oh no: %s" % e)
		except socket.sslerror, e:
			self._clear_connection()
			if e.__str__().find("timed out") != -1: # Connection timeouts should be mapped to RemoteExceptions
				raise RemoteException(self.host, e)
			else:
				raise RGMASecurityException("Secure connection failed: %s" % e, 0, e)
		except socket.error, e:
			self._clear_connection()
			raise RemoteException(self.host, e)
		except AssertionError, e:
			# This is a workaround for python bug #666219 which can cause AssertionError to be thrown
			# when the server closes unexpectedly.
			self._clear_connection()
			raise RemoteException(self.host,"Assertion Error caught in connect - It looks like this is the buggy version of Python - please consider using a different version.It has been tested with Python 2.4.1 %s" % e)
	def _sendMessage(self, path, command, encodedParams, headers):
		
		try:
			if _use_alarm:
				
				signal.alarm(_connection_timeout)
			self.conn.request("GET", "%s/%s?%s" % (self.path, command, encodedParams), None, headers)
		       	return self.conn.getresponse()
	       	finally:
		       	if _use_alarm:
				
			       	signal.alarm(0) # Clear the previously set alarm
					
						
	def _callServlet(self, command, **params):
		for param in params.keys():
			if params[param] == None:
				del params[param]

		try:
			if self.conn == None:
				self._connect()
			headers = {"Cache-Control" : "no-cache",
				   "Connection" : "keep-alive"}
		
			try:
				
				resp = self._sendMessage(self.path, command, _urlencode(params), headers)
				if resp.status != 200:
					raise RemoteException(self.host, "HTTP error %i (%s)" % (resp.status, resp.reason))
				xmlData = resp.read()
			
			except(httplib.HTTPException, AssertionError), e:
				
				#Its possible the connection has died and will need to be re-established
				self._clear_connection()
				self._connect()
				resp = self._sendMessage(self.path, command, _urlencode(params), headers)
				if resp.status != 200:
					raise RemoteException(self.host, "HTTP error %i (%s)" % (resp.status, resp.reason))
				xmlData = resp.read()
			
			return xmlData
		
		except IOError, e:
			self._clear_connection()
			raise RemoteException(self.host, e.strerror)
		except httplib.HTTPException, e:
			self._clear_connection()
			raise RemoteException(self.host, "HTTP connection failed: %s" % e)
		except socket.sslerror, e:
			self._clear_connection()
			if e.__str__().find("timed out") != -1: # Connection timeouts should be mapped to RemoteExceptions
				raise RemoteException(self.host, e)
			else:
				raise RGMASecurityException("Secure connection failed: %s" % e, 0, e)
		except socket.error, e:
			self._clear_connection()
			raise RemoteException(self.host, e)
		except AssertionError, e:
			# This is a workaround for python bug #666219 which can cause AssertionError to be thrown
			# when the server closes unexpectedly.
			self._clear_connection()
			raise RemoteException(self.host,"Assertion Error caught when attempting to send message - It looks like this is the buggy version of Python - please consider using a different version %s" % e )

	def callServiceMethod(self, command, returnType=None, **params):
		"""
		Call a method on the service.

		Parameters:

		  command    - Name of the method to call
		  returnType - Type of the returned object (default None)
		  params     - key/value parameters.

		If returnType is specified, the returned ResultSet will be converted
		to this type of object, if possible. Integers and boolean values
		are assumed to be contained in the first column of the first row
		of the ResultSet. Other objects will have their constructors called
		with the arguments 'resultSet' and 'fromUrl' set to appropriate values. 
		"""
		xmlData = self._callServlet(command, **params)
		rs = _ResponseParser(xmlData).resultSet
		
		try:
			if returnType == int:
				return int(rs[0][0])
			elif returnType == bool:
				return rs[0][0].lower() == "true"
			elif returnType == str:
				return rs[0][0]
			elif returnType == ResultSet:
				return rs
			elif returnType:
				return returnType(resultSet=rs, fromUrl=self.url)
		except Exception, e:
			raise RGMAException("Invalid response from service", 0, e) 

	def boolToString(self, boolValue):
		"""
		Encodes boolean 0/1 value as the strings "false"/"true", as
		reqired by the services.

		Parameters
		  boolValue    - Boolean value to encode.
		"""
		if (boolValue):
			return "true"
		else:
			return "false"

	def getVersion(self):
		"""
		Get the service version
		
		Returns a version string in the form major.minor.patch
		"""
		return self.callServiceMethod("getVersion", str)

	def getProperty(self, name, parameter):
		"""
		Get a service property

		Parameters:
		  name - Name of the property to get
		  parameter - Optional parameter (may be None)
		"""
		if parameter == None:
			return self.callServiceMethod("getProperty", str,
						      name=name)
		else:
			return self.callServiceMethod("getProperty", str,
						      name=name,
						      parameter=parameter)

	def setProperty(self, name, parameter, value):
		"""
		Set a service property

		Parameters:
		  name - Name of the property
		  parameter - Optional parameter (may be None)
		  value    - Value to set
		"""
		if parameter == None:
			return self.callServiceMethod("setProperty", str,
						      name=name,
						      value=value)
		else:
			return self.callServiceMethod("setProperty", str,
						      name=name,
						      parameter=parameter,
						      value=value)

class Resource(Service):
	"""
	Base class for a connection to a soft-state RGMA resource (e.g. Consumer, 
	Producer, but not a Schema or Registry.)
	
	A resource is identified by a ResourceEndpoint (a URL location with an
	additional context ID). It has a 'termination interval' which specifies
	how long the resource will remain alive if no contact is made with it.
	Contact can be made by calling a method on the resource, or using the
	showSignOfLife method.

	Resources can also be closed explicitly, in which case they will finish
	off any outstanding tasks before dying, or destroyed, in which case they
	will die immediately.

	properties

	  endpoint - ResourceEndpoint of the resource
	"""

	def __init__(self, serviceName, endpoint):
		"""
		Constructor

		Parameters:

		  serviceName - Service name in rgma.conf (e.g. "Consumer")
		  endpoint - ResourceEndpoint of the resource
		"""
		self.endpoint = endpoint
		Service.__init__(self, serviceName, endpoint.url)
		
	def callMethod(self, method, returnType=None, **params):
		"""
		Call a resource method.

		Parameters:

		  method     - Name of the method to call
		  returnType - Type of the returned object (default None)
		  params     - key/value parameters.

		The resource connection ID is added to the service call automatically.
		Service.callServiceMethod is then invoked with all other parameters.		
		"""
		params['connectionId'] = self.endpoint.resourceId
		return self.callServiceMethod(method, returnType, **params)
			
	def close(self):
		"""
		Closes the resource. 
		
		The resource will no longer be available through this API and 
		will be destroyed once it has finished interacting with other 
		components.
		"""
		self.callMethod("close")
					
	def destroy(self):
		"""
		Close the resource immediately
		
		The resource will no longer be available to any component.
		"""
		self.callMethod("destroy")
		
	def showSignOfLife(self):
		"""
		Indicate to the resource that this connection is still active.
		
		This should be used to keep the resource alive when no
		other methods need to be called.
		"""
		self.callMethod("showSignOfLife")
			
	def setTerminationInterval(self, interval):
		"""
		Set the termination interval for this resource

		THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE VERSION
		
		Patemeters:
		
  		  interval - TimeInterval object specifying the length of time the
		             resource will stay active if no connections are made 
		             to it. After this time, the resource will close 
		             automatically.
		"""
		self.callMethod("setTerminationInterval",
				 terminationIntervalSec=int(interval.getValueAs(Units.SECONDS)))
			
	def getEndpoint(self):
		"""
		Return the ResourceEndpoint of the resource
		"""
		return self.endpoint

	def getTerminationInterval(self):
		"""
		Return the termination interval for this resource as a 
		TimeInterval object.
		"""
		return self.callMethod("getTerminationInterval",
					TimeInterval)
	
def reconnect(resource, endpoint):
	"""
	Reconnect to an existing RGMA resource
	
	Parameters:
	
	resource - Resource to reconnect to. Must be one of
	             rgma.Consumer
	             rgma.PrimaryProducer
	             rgma.SecondaryProducer
	             rgma.OnDemandProducer
	endpoint - ResourceEndpoint object containing the URL and resource ID
	           of the existing resource.
	"""
	res = resource(None, None, None, endpoint=endpoint)
	return res
	
class Consumer(Resource):
	"""
	A Consumer is used to query R-GMA for information
	
	Consumers execute an SQL query over the R-GMA information system. The
	query may be over the latest tuples for each primary key, historical
	tuples or may be 'continuous' in which case matching tuples are returned
	as they are inserted (while the Consumer remains alive).
	
	The query will only be executed once the start method has been called.
	Matching tuples are retrieved in groups with a maximum size. When the
	query has completed, either because all results have been returned or because
	the query times out, a flag is set on the returned ResultSet.
	"""
	
	def __init__(self, terminationInterval, queryProperties, \
	             selectStatement, **kwargs):
		"""
		Constructor
		
		Parameters:
		
		terminationInterval - TimeInterval specifying the termination 
		                      interval for the resource
		queryProperties     - A QueryProperties object
		selectStatement     - SQL SELECT query to execute
		"""
		if kwargs.has_key('endpoint'):
			Resource.__init__(self, "Consumer", kwargs['endpoint'])
		else:
			if kwargs.has_key('serviceUrl'):
				url = kwargs['serviceUrl']
			else:
				url = None
			Service.__init__(self, "Consumer", url)

			self.queryProperties = queryProperties
			self.terminationInterval = terminationInterval
			self.select = selectStatement
			if queryProperties.hasTimeInterval():
				timeIntervalSec = int(queryProperties.timeInterval.getValueAs(Units.SECONDS))
			else:
				timeIntervalSec = None
			endpoint = self.callServiceMethod("createConsumer",
							  ResourceEndpoint,
							  terminationIntervalSec=int(terminationInterval.getValueAs(Units.SECONDS)),
							  select=selectStatement,
							  isLatest=self.boolToString(queryProperties.isLatest()),
							  isHistory=self.boolToString(queryProperties.isHistory()),
							  isContinuous=self.boolToString(queryProperties.isContinuous()),
							  isStatic=self.boolToString(queryProperties.isStatic()),
							  timeIntervalSec=timeIntervalSec)
			Resource.__init__(self, "Consumer", endpoint)
			
	def abort(self):
		"""
		Abort the query if it is executing
		"""
		self.callMethod("abort")

	def count(self):
		"""
		Return the number of tuples available to pop

		THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE VERSION
		"""
		return self.callMethod("count", int)

	def hasAborted(self):
		"""
		Return True if the query has been aborted
		"""
		return self.callMethod("hasAborted", bool)

	def isExecuting(self):
		"""
		Return True if the query is currently executing

		THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE VERSION
		"""
		return self.callMethod("isExecuting", bool)
	
	def pop(self, maxCount=None):
		"""
		Return a ResultSet containing tuples that are the results of
		the query.
		
		Parameters:
		
		maxCount - Maximum number of tuples to return.

		If maxCount is not specified, all available tuples will be returned.
		THIS IS DEPRECATED AND SHOULD NOT BE USED - ALWAYS SPECIFY A MAXIMUM COUNT.
		"""
		if maxCount:
			return self.callMethod("pop", ResultSet, maxCount=maxCount)
		else:
			return self.callMethod("pop", ResultSet)
		
	def start(self, timeout=None, producers=None):
		"""
		Start executing the query

		Parameters:

		timeout   - TimeInterval specifying how long to run the query for before
		            automatically aborting (optional - default no limit).
		producers - List of ResourceEndpoint objects defining the producers to
		            contact to answer the query. If specified R-GMA will bypass
			    its usual mediation process and query the specified producers
			    only. It is the user's responsibility to ensure that the
			    producers are able to answer the specified query type.
		"""
		if timeout:
			timeout = int(timeout.getValueAs(Units.SECONDS))

		if producers:
			endpoints = []
			for producer in producers:
				endpoints.append("%i %s" % (producer.resourceId, producer.url))
			self.callMethod("startDirected", producerConnections=endpoints, timeoutSec=timeout)
		else:
			self.callMethod("start", timeoutSec=timeout)
		
	def getVersion():
		service = Service("Consumer", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the Consumer service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("Consumer", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the Consumer service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("Consumer", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the Consumer service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """

	def __str__(self):
		return "Consumer at %s: %s '%s' (termination interval %s)" % \
		       (self.endpoint, self.queryProperties, self.select, self.terminationInterval)
	
class PrimaryProducer(Resource):
	"""
	A Primary Producer is used to put new information into the
	R-GMA system.
	
	Tuples are inserted into named tables using standard SQL INSERT
	statements. The tables must already exist in the global schema, and
	before publishing to a table a primary producer must call its
	declareTable method to announce its intention.
	"""
	
	def __init__(self, terminationInterval, producerProperties, \
	             voNames=None, **kwargs):
		"""
		Constructor
		
		Parameters:
		
		terminationInterval - TimeInterval object specifying the 
		                      termination interval for the resource
		producerProperties  - A ProducerProperties object
		voNames             - A list of Virtual Organisations to publish 
		                      in (DEPRECATED - DO NOT USE)
		"""
		if kwargs.has_key('endpoint'):
			Resource.__init__(self, "PrimaryProducer", kwargs['endpoint'])
		else:
				
			if kwargs.has_key('serviceUrl'):
				url = kwargs['serviceUrl']
			else:
				url = None
			Service.__init__(self, "PrimaryProducer", url)

			self.terminationInterval = terminationInterval
			self.producerProperties = producerProperties
			storage = producerProperties.storage
			location = None
			username = None
			password = None
			logicalName = storage.logicalName
			if storage.username and storage.password:
				location = storage.location
				username = storage.username
				password = storage.password
				logicalName = None

			endpoint = self.callServiceMethod("createPrimaryProducer",
							  ResourceEndpoint,
							  terminationIntervalSec=int(terminationInterval.getValueAs(Units.SECONDS)),
							  type=StorageType.TypeName[producerProperties.storage.type],
							  storageLocation=location,
							  storageUserName=username,
							  storagePassword=password,
							  storageLogicalName=logicalName,
							  isLatest=self.boolToString(producerProperties.isLatest()),
							  isHistory=self.boolToString(producerProperties.isHistory()))
			Resource.__init__(self, "PrimaryProducer", endpoint)
			
	def declareTable(self, tableName, predicate, \
	                 historyRetentionPeriod, \
	                 latestRetentionPeriod):
		"""
		Declare a table that this Producer may publish tuples into
		
		Parameters:
		
		  tableName              - The table to publish into
		  predicate              - An SQL WHERE predicate which must be
		                           satisfied by every tuple this Producer
		                           publishes. May be an empty string if no
			  		   predicate is required.
		  historyRetentionPeriod - TimeInterval object specifying the 
		                           length of time to keep tuples 
		                           available to History queries.
		  latestRetentionPeriod  - TimeInterval object specifying the default 
		                           length of time to keep tuples
		                           available to Latest queries. This may be
					   overridden in the insert method.
		"""
		self.callMethod("declareTable", tableName=tableName,
				 predicate=predicate,
				 hrpSec=int(historyRetentionPeriod.getValueAs(Units.SECONDS)),
				 lrpSec=int(latestRetentionPeriod.getValueAs(Units.SECONDS)))

	def getHistoryRetentionPeriod(self, tableName):
		"""
		Get the history retention period of a given table
		
		This is a TimeInterval specifying the minimum length of time 
		that tuples will be kept in the Producer's History queue

		Parameters:

		  tableName - The name of the table to return the retention period for
		"""
		return self.callMethod("getHistoryRetentionPeriod", TimeInterval, tableName=tableName)
		
	def getLatestRetentionPeriod(self, tableName):
		"""
		Get the latest retention period of a given table
		
		This is a TimeInterval specifying the minimum length of time 
		that tuples will be kept in the Producer's Latest queue

		Parameters:

		  tableName - The name of the table to return the retention period for
		"""
		return self.callMethod("getLatestRetentionPeriod", TimeInterval, tableName=tableName)
	
	def insert(self, insertStatement, latestRetentionPeriod=None):
		"""
		Insert a tuple into a table
		
		The table must have been declared with declareTable. 
		
		Parameters:
		
  		  insertStatement       - SQL INSERT statement
		  latestRetentionPeriod - TimeInterval specifying how long this tuple
		                          should be considered as the latest information
					  for its primary key. If not specified, the default
					  given in declareTable is used.
		"""
		if latestRetentionPeriod:
			lrpSec = int(latestRetentionPeriod.getValueAs(Units.SECONDS))
		else:
			lrpSec = None
		self.callMethod("insert", insert=insertStatement, lrpSec=lrpSec)
	
	def insertList(self, insertStatementList, latestRetentionPeriod=None):
		"""
		Insert a list of tuples
			
		All tables inserted to must have been declared with declareTable
		
		Parameters:
		
  		  insertStatementList   - A list of SQL INSERT statements
		  latestRetentionPeriod - TimeInterval specifying how long these tuples
		                          should be considered as the latest information
					  for their primary keys. If not specified, the default
					  given in declareTable is used.
		"""
		if latestRetentionPeriod:
			lrpSec = int(latestRetentionPeriod.getValueAs(Units.SECONDS))
		else:
			lrpSec = None
		self.callMethod("insertList", insert=insertStatementList, lrpSec=lrpSec)

	def listTupleStores():
		service = Service("PrimaryProducer", None)
		rs = service.callServiceMethod("listTupleStores", ResultSet)
		stores = []
		for store in rs:
			stores.append(TupleStore(StorageType.Type[store[0].lower()], store[1], store[2]))
		return stores
	listTupleStores = staticmethod(listTupleStores)
	
	def dropTupleStore(logicalName):
		service = Service("PrimaryProducer", None)
		rs = service.callServiceMethod("dropTupleStore", logicalName=logicalName)
	dropTupleStore = staticmethod(dropTupleStore)
		
	def getVersion():
		service = Service("PrimaryProducer", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the PrimaryProducer service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("PrimaryProducer", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the PrimaryProducer service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("PrimaryProducer", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the PrimaryProducer service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """
	
	def __str__(self):
		return "Primary Producer at %s" % self.endpoint
	
class SecondaryProducer(Resource):
	"""
	A SecondaryProducer collects information from PrimaryProducers
	and republishes it to R-GMA.
	
	SecondaryProducers are used to bring tuples from different
	PrimaryProducers together so that join queries may be performed. They
	may also be used to archive data from PrimaryProducers that have
	do not support history/latest queries, or have short retention periods.
	"""
	
	def __init__(self, terminationInterval=None, producerProperties=None, \
	             voNames=None, **kwargs):
		"""
		Constructor
		
		Parameters:
		
		terminationInterval - TimeInterval object specifying the 
		                      termination interval for the resource
		producerProperties  - A ProducerProperties object
		voNames             - A list of Virtual Organisations to publish
		                      to (DEPRECATED - DO NOT USE)
		"""
		if kwargs.has_key('endpoint'):
			Resource.__init__(self, "SecondaryProducer", kwargs['endpoint'])
		else:
			if not terminationInterval:
				raise RGMAException("termination interval cannot be None")
			if not producerProperties:
				raise RGMAException("producer properties cannot be None")

			if kwargs.has_key('serviceUrl'):
				url = kwargs['serviceUrl']
			else:
				url = None
			Service.__init__(self, "SecondaryProducer", url)
				
			self.terminationInterval = terminationInterval
			self.producerProperties = producerProperties
			storage = producerProperties.storage
			location = None
			username = None
			password = None
			logicalName = storage.logicalName
			if storage.username != None and storage.password != None:
				location = storage.location
				username = storage.username
				password = storage.password
				logicalName = None
			endpoint = self.callServiceMethod("createSecondaryProducer",
							  ResourceEndpoint,
							  terminationIntervalSec=int(terminationInterval.getValueAs(Units.SECONDS)),
							  type=StorageType.TypeName[producerProperties.storage.type],
							  storageLocation=location,
							  storageUserName=username,
							  storagePassword=password,
							  storageLogicalName=logicalName,
							  isLatest=self.boolToString(producerProperties.isLatest()),
							  isHistory=self.boolToString(producerProperties.isHistory()))
			Resource.__init__(self, "SecondaryProducer", endpoint)

	def declareTable(self, tableName, predicate, \
	                 historyRetentionPeriod):
		"""
		Declare a table that this SecondaryProducer republishes
		
		The latest retention period is inherited from the individual
		PrimaryProducers the original data comes from.
		
		Parameters:
		
		tableName              - The table to republish
		predicate              - The SQL WHERE predicate giving the part
		                         of the table to republish
		historyRetentionPeriod - TimeInterval specifying the length of 
		                         time to keep tuples available to 
		                         History queries.
		"""
		self.callMethod("declareTable", tableName=tableName,
				 predicate=predicate, hrpSec=int(historyRetentionPeriod.getValueAs(Units.SECONDS)))
		
	def getHistoryRetentionPeriod(self, tableName):
		"""
		Get the history retention period for the SecondaryProducer
		
		This is a TimeInterval specifying the minimum length of time 
		that tuples will be kept in the SecondaryProducer's History 
		queue

		Parameters:

		tableName - The name of the table to return the retention period for
		"""		
		return self.callMethod("getHistoryRetentionPeriod", TimeInterval, tableName=tableName)

	def listTupleStores():
		service = Service("SecondaryProducer", None)
		rs = service.callServiceMethod("listTupleStores", ResultSet)
		stores = []
		for store in rs:
			stores.append(TupleStore(StorageType.Type[store[0].lower()], store[1], store[2]))
		return stores
	listTupleStores = staticmethod(listTupleStores)

	def dropTupleStore(logicalName):
		service = Service("SecondaryProducer", None)
		rs = service.callServiceMethod("dropTupleStore", logicalName=logicalName)
	dropTupleStore = staticmethod(dropTupleStore)
		
	def getVersion():
		service = Service("SecondaryProducer", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the SecondaryProducer service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("SecondaryProducer", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the SecondaryProducer service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("SecondaryProducer", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the SecondaryProducer service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """

	def __str__(self):
		return "Secondary Producer at %s" % self.endpoint
	
class OnDemandProducer(Resource):
	"""
	An OnDemandProducer is used to publish information into R-GMA in
	response to specific Consumer queries.
	
	An OnDemandProducer may be used to make an existing data store (normally
	a database) available for querying through R-GMA. It may also be useful
	if the cost of creating the data to answer a query is high since with an
	OnDemandProducer, this only needs to be done in response to specific Consumer
	queries.
	
	OnDemandProducers answer a special 'static' query type. They cannot answer
	continuous, latest or history queries. When a static query is issued it
	is forwarded from the OnDemandProducer service to a user-supplied socket://
	URL. The user code listening at this URL is responsible for generating the
	result set which answers the query and returning it to the OnDemandProducer
	service. The OnDemandProducer service is responsible for streaming the data
	to the Consumer.

	The OnDemandProducer sends the unencoded SELECT statement to the socket using
	the hostname and port specified in the URL. It reads the response back using
	the same socket connection.
	"""
	
	def __init__(self, terminationInterval=None, uri=None, voNames=None, **kwargs):
		"""
		Constructor
		
		Parameters:
		
		terminationInterval - TimeInterval object specifying the 
		                      termination interval for the resource
		uri                 - A socket:// URL where requests from the OnDemandProducer
		                      service should be sent to user code.
		voNames             - A list of Virtual Organisations to publish
		                      to (DEPRECATED - DO NOT USE)
		"""
		if kwargs.has_key('endpoint'):
			Resource.__init__(self, "OnDemandProducer", kwargs['endpoint'])
		else:
			if not terminationInterval:
				raise RGMAException("termination interval cannot be None")
			if not uri:
				raise RGMAException("producer properties cannot be None")

			if kwargs.has_key('serviceUrl'):
				url = kwargs['serviceUrl']
			else:
				url = None
			Service.__init__(self, "OnDemandProducer", url)
				
			self.terminationInterval = terminationInterval
			self.uri = uri
			endpoint = self.callServiceMethod("createOnDemandProducer",
							  ResourceEndpoint,
							  terminationIntervalSec=int(terminationInterval.getValueAs(Units.SECONDS)),
							  uri=uri)
			Resource.__init__(self, "OnDemandProducer", endpoint)
			
	def declareStaticTable(self, tableName, predicate):
		"""
		Declare a static table that this OnDemandProducer publishes to
		
		Parameters:
		
		tableName - The table to republish
		predicate - An SQL WHERE predicate which all tuples from this
		            producer must satisfy
		"""
		self.callMethod("declareStaticTable",
				 tableName=tableName, predicate=predicate)
		
	def getVersion():
		service = Service("OnDemandProducer", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the OnDemandProducer service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("OnDemandProducer", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the OnDemandProducer service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("OnDemandProducer", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the OnDemandProducer service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """
			
	def __str__(self):
		return "On-Demand Producer at %s" % self.endpoint

class ProducerDescription:
	"""
	Description of a Producer.

	properties:

	  queryTypes        - Query types supported - see QueryTypes
	  endpoint          - ResourceEndpoint of producer
	  hrp               - Producer history retention period
	  secondaryProducer - Whether the producer is a secondary producer
	  qosAttribute      - Quality of service attribute
	"""

	def __init__(self, endpoint, queryTypes, secondaryProducer, hrp, qosAttribute=None):
		self.endpoint = endpoint
		self.queryTypes = queryTypes
		self.secondaryProducer = secondaryProducer
		self.hrp = hrp
		self.qosAttribute = qosAttribute

	def getQueryTypes(self):
		"""
		Get the query types supported as a QueryType integer
		"""
		return self.queryTypes

	def getEndpoint(self):
		"""
		Get the ResourceEndpoint of the producer
		"""
		return self.endpoint

	def getHistoryRetentionPeriod(self):
		"""
		Get the History retention period of the producer as a TimeInterval
		"""
		return self.hrp
	
	def isRepublisher(self):
		"""
		Returns true if the producer is part of a
		SecondaryProducer
		"""
		return self.secondaryProducer

	def isSecondaryProducer(self):
		"""
		Returns true if the producer is part of a
		SecondaryProducer
		"""
		return self.secondaryProducer
	
	def getQosAttribute(self):
		"""
		Get the Quality of Service attribute
		"""
		return self.qosAttribute

	def __str__(self):
		if self.queryTypes == QueryType.STATIC:
			str = "On-Demand Producer at "
		else:
			str = QueryType.TypeName[self.queryTypes].upper()
			if self.secondaryProducer:
				str += " Secondary Producer at "
			else:
				str += " Primary Producer at "
			
		return str + "%s" % self.endpoint

class Index:
        """
	Information about an index on a table

	properties:

	  name    - Name of the index
	  columns - List of column names to be indexed on
	"""

	def __init__(self, name, columns):
		self.name = str(name)
		self.columns = columns

	def __str__(self):
		columns = string.join(self.columns, " ")
		return "%s [%s]" % (self.name, columns)
		
class TableDefinition:
	"""
	Information about a table from the Schema
	"""
	def __init__(self, tableName, columns=[]):
		self.name = tableName
		self.columns = columns

	def __str__(self):
		str = "%s (" % self.name
		comma = ''
		for column in self.columns:
			str += "%s%s" % (comma, column)
			comma = ', '
		return str + ")"
		
class ColumnDefinition:
	"""
	Information about a column

	properties:

	  name       - Column name
	  type       - Column type (see Types)
	  width      - Column width (e.g. for VARCHAR(240), 240 is the width), or
	               None if there is no width
	  notNull    - If true, column cannot contain a null value
	  primaryKey - If true, column is part of the primary key
	"""

	def __init__(self, name, type, width=None, notNull=False, primaryKey=False):
		self.name = name
		self.type = type
		self.width = width
		self.notNull = notNull
		self.primaryKey = primaryKey

	def __str__(self):
		str = "%s %s" % (self.name, TypeString(self.type, self.width))
		if self.primaryKey: str += " PRIMARY KEY"
		if self.notNull: str += " NOT NULL"
		return str
	
class Registry(Service):
	"""
	Class providing an interface to the Registry for a virtual database.

	Each virtual database (VDB) has its own separate logical Registry (which
	may, however, be replicated by multiple registry services). This registry
	contains information about all the producers and consumers publishing
	or querying tables in the VDB.

	A Registry Service is a stateless R-GMA service which processes
	requests for all VDB Registries. It may host any number of Registry Instances for
	particular VDBs. If a registry service receives a request for a VDB it
	does not host locally it forwards it on to another registry service which does.
	A registry service which does not host any VDB instances is purely a proxy
	service.
	"""
	def __init__(self, vdbName=None):
		"""
		Constructor

		Parameters:
		  vdbName - Name of the Virtual Database to connect to
		"""
		self.vdbName = vdbName
		Service.__init__(self, "Registry")

	def createRegistry(self, registryService):
		"""
		Create a Registry instance for this virtual database on
		a specified registry service

		Parameters:

		  registryService - URL of a Registry service
		"""
		self.callServiceMethod("createRegistry",
				       vdbName=self.vdbName,
				       registryService=registryService)

	def destroyRegistry(self, registryService):
		"""
		Destroy the Registry instance for this virtual database on
		a specified registry service

		Parameters:

		  registryService - URL of a Registry service
		"""
		self.callServiceMethod("destroyRegistry",
				       vdbName=self.vdbName,
				       registryService=registryService)

	def getAllProducersForTable(self, tableName):
		"""
		Get a list of producers that publish a table in this virtual
		database

		Returns a list of ProducerDescription objects

		Parameters:
		  tableName  - Name of table
		"""
		rs = self.callServiceMethod("getAllProducersForTable",
					    ResultSet,
					    vdbName=self.vdbName,
					    tableName=tableName,
					    canForward="true")
		producers = []
		for producer in rs:
			queryType = 0
			if producer['isHistory'].lower() == "true": queryType = queryType | QueryType.HISTORY
			if producer['isLatest'].lower() == "true": queryType = queryType | QueryType.LATEST
			if producer['isContinuous'].lower() == "true": queryType = queryType | QueryType.CONTINUOUS
			if producer['isStatic'].lower() == "true": queryType = queryType | QueryType.STATIC
			producers.append(ProducerDescription(ResourceEndpoint(producer['url'], producer['connectionId']), queryType, producer['isSecondaryProducer'].lower() == "true", TimeInterval(producer['hrpSec'], Units.SECONDS), producer['qosAttrib']))
				
		return producers
		
	def getVersion():
		service = Service("Registry", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the Registry service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("Registry", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the Registry service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("Registry", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the Registry service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """

	def __str__(self):
		str = "Registry at %s" % self.url
		if self.vdbName:
			str += " for virtual database '%s'" % self.vdbName
		return str

class Schema(Service):
	"""
	Class providing an interface to the Schema for a virtual
	database

	Each virtual database (VDB) has a Schema which defines the tables
	in the VDB and the authorisation rules for publishing and querying
	the tables.

	A Schema Service is a stateless services that processes request for
	all VDB schemas. It may host any number of VDB Schema Instances in which
	case it can answer the request directly. Otherwise it will forward the request
	to another Schema Service which does host the relevant VDB. A Schema service
	which does not host any VDB instances is a pure proxy service.
	"""
	def __init__(self, vdbName=None):
		"""
		Constructor

		Parameters

		  vdbName - Name of the virtual database to connect to
	        """
		self.vdbName = vdbName
		Service.__init__(self, "Schema")

	def createSchema(self, schemaService):
		"""
		Create a Schema instance for this virtual database on
		a specified schema service

		Parameters:

		  schemaService - URL of a Schema service
		"""
		self.callServiceMethod("createSchema",
				       vdbName=self.vdbName,
				       schemaService=schemaService)

	def destroySchema(self, schemaService):
		"""
		Destroy the Schema instance for this virtual database on
		a specified schema service

		Parameters:

		  schemaService - URL of a Schema service
		"""
		self.callServiceMethod("destroySchema",
				       vdbName=self.vdbName,
				       schemaService=schemaService)

	def createTable(self, createTableStatement, tableAuthz):
		"""
		Create a table in the virtual database

		Parameters:

		  createTableStatement - SQL 'CREATE TABLE' statement
		  tableAuthz           - a list of table authorisation rules
		"""
		self.callServiceMethod("createTable",
				       vdbName=self.vdbName,
				       canForward="true",
				       createTableStatement=createTableStatement,
				       tableAuthz=tableAuthz)

	def dropTable(self, tableName):
		"""
		Drop a table from the virtual database

		Parameters:

		  tableName - Name of the table to drop
		"""
		self.callServiceMethod("dropTable",
				       vdbName=self.vdbName,
				       canForward="true",
				       tableName=tableName)

	def createIndex(self, createIndexStatement):
		"""
		Create an index in the virtual database

		Parameters:

		  createIndexStatement - SQL 'CREATE INDEX' statement
		"""
		self.callServiceMethod("createIndex",
				       vdbName=self.vdbName,
				       canForward="true",
				       createIndexStatement=createIndexStatement)

	def dropIndex(self, indexName):
		"""
		Drop an index from the virtual database

		Parameters:

		  indexName - Name of the index to drop
		"""
		self.callServiceMethod("dropIndex",
				       vdbName=self.vdbName,
				       canForward="true",
				       indexName=indexName)

	def createView(self, createViewStatement, viewAuthz):
		"""
		Create a view in the virtual database

		Parameters:

		  createViewStatement - SQL 'CREATE VIEW' statement
		  viewAuthz           - list of authorisation rules
		"""
		self.callServiceMethod("createView",
				       vdbName=self.vdbName,
				       canForward="true",
				       createViewStatement=createViewStatement,
				       viewAuthz=viewAuthz)

	def dropView(self, viewName):
		"""
		Drop a view from the virtual database

		Parameters:

		  viewName - Name of the view to drop
		"""
		self.callServiceMethod("dropView",
				       vdbName=self.vdbName,
				       canForward="true",
				       viewName=viewName)

	def getAllTables(self):
		"""
		Returns a list of names of tables defined in the virtual database
		"""
		rs = self.callServiceMethod("getAllTables", ResultSet, vdbName=self.vdbName, canForward="true")
		tables = []
		for table in rs: tables.append(table['name'])
		return tables

	def getTableDefinition(self, tableName):
		"""
		Get the definition of a table in the virtual database

		Parameters:

		  tableName - Name of the table to query

		Returns a TableDefinition object
		"""
		rs = self.callServiceMethod("getTableDefinition", ResultSet,
					    tableName=tableName, vdbName=self.vdbName, canForward="true")
		columns = []
		for column in rs:
			try:
				type = Types.Type[column['type'].upper()]
			except KeyError:
				type = Types.UNKNOWN
			notNull = column['notNull'].strip().lower() == "true"
			primaryKey = column['primaryKey'].strip().lower() == "true"
			width = int(column['size'])
			if width == 0: width = None
			columns.append(ColumnDefinition(column['columnName'], type, width, notNull, primaryKey))
				       
		return TableDefinition(tableName, columns)

	def getTableIndexes(self, tableName):
		"""
		Get a list of indexes associated with a table

		Parameters:

		  tableName - Name of the table. Indexes cannot be associated with a view
		"""
		rs = self.callServiceMethod("getTableIndexes", ResultSet,
					    tableName=tableName, vdbName=self.vdbName, canForward="true")
		indexColumns = {}
		for index in rs:
			indexName = index['indexName']
			if not indexColumns.has_key(indexName):
				indexColumns[indexName] = []
				
			indexColumns[indexName].append(index['columnName'])

		indexes = []
		for indexName in indexColumns:
			indexes.append(Index(indexName, indexColumns[indexName]))

		return indexes
	
	def setAuthorizationRules(self, tableName, tableAuthz):
		"""
		Set the authorisation rules for a table or view
		
		Parameters:

		  tableName  - Name of the table or view
		  tableAuthz - List of authorisation rules
		"""
		self.callServiceMethod("setAuthorizationRules",
				       tableName=tableName,
				       tableAuthz=tableAuthz,
				       vdbName=self.vdbName,
				       canForward="true")

	def getAuthorizationRules(self, tableName):
		"""
		Get the authorisation rules for a table or view

		Returns a list of rules.
		
		Parameters:

		  tableName  - Name of the table or view
		"""
		rs = self.callServiceMethod("getAuthorizationRules", ResultSet,
					    tableName=tableName, vdbName=self.vdbName, canForward="true")
		rules = []
		for rule in rs:
			rules.append(rule['rule'])
		return rules
				
	def getVersion():
		service = Service("Schema", None)
		return service.getVersion()

	getVersion = staticmethod(getVersion)
	getVersion.__doc__ = """
	                     Get the version of the Schema service
			     """
	
	def getProperty(name, parameter=None):
		service = Service("Schema", None)
		return service.getProperty(name, parameter)
	
	getProperty = staticmethod(getProperty)
	getProperty.__doc__ = """
	                      Get a property of the Schema service

			      Parameters:

			      name      - Name of property
			      parameter - Optional parameter
			      """
	
	def setProperty(name, value, parameter=None):
		service = Service("Schema", None)
		return service.setProperty(name, parameter, value)
	
	setProperty = staticmethod(setProperty)
	setProperty.__doc__ = """
	                      Set a property of the Schema service

			      Parameters:

			      name      - Name of property
			      value     - Value to set
			      parameter - Optional parameter
			      """

	def __str__(self):
		str = "Schema at %s" % self.url
		if self.vdbName:
			str += " for virtual database '%s'" % self.vdbName
		return str
