"""
Tests for the R-GMA Python API
"""

import os, time, random, md5
import unittest

import rgma

class TestBase:

	TEST_TABLE = "userTable"
	TEST_ID_COLUMN = "userId"
	TEST_STRING_COLUMN = "aString"
	TEST_REAL_COLUMN = "aReal"
	TEST_INT_COLUMN = "anInt"
	DATE_COLUMN = "MeasurementDate"
	TIME_COLUMN = "MeasurementTime"
	
        def __init__(self):
            try:
                rgmaHome = os.environ['RGMA_HOME']
            except KeyError:
                raise "ERROR: Environment variable RGMA_HOME is not set"
            
            self.createUniqueId()
            
        def createUniqueId(self, args=None):
            """
            Create a unique ID to insert into userTable. This ensures that
            tests work if there is other data in the table
            """
            t = long( time.time() * 1000 )
            r = long( random.random()*100000000000000000L )
            try:
                a = socket.gethostbyname( socket.gethostname() )
            except:
                # if we can't get a network address, just imagine one
                a = random.random()*100000000000000000L
            data = str(t)+' '+str(r)+' '+str(a)+' '+ str(args)
            self.uniqueId = md5.md5(data).hexdigest()

	def insertTestTuple(self):
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		
class ConsumerTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(ConsumerTest,'test')
		
	def testCreateContinuous(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.close()
		
	def testCreateLatest(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.LATEST), query)
		consumer.close()

	def testCreateHistory(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.HISTORY), query)
		consumer.close()
		
	def testCreateContinuousInterval(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		interval = rgma.TimeInterval(1, rgma.Units.DAYS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS, interval), query)
		consumer.close()
		
	def testCreateLatestInterval(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		interval = rgma.TimeInterval(1, rgma.Units.DAYS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.LATEST, interval), query)
		consumer.close()

	def testCreateHistoryInterval(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		interval = rgma.TimeInterval(1, rgma.Units.DAYS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.HISTORY, interval), query)
		consumer.close()

	def testIsExecuting(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(2, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		assert(not consumer.isExecuting())
		consumer.start(timeout)
		assert(consumer.isExecuting())

	def testStartTimeout(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(2, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)
		time.sleep(3)
		assert(not consumer.isExecuting())

	def testStart(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start()
		time.sleep(1)
		assert(consumer.isExecuting())
		consumer.abort()
		consumer.close()

	def testStartDirected(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(5, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)

		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
 		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.PrimaryProducer(ti, pp)
 		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		
		re = p.getEndpoint()
		
		consumer.start(timeout, [re,])
		
	def testAbort(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(5, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)
		consumer.abort()
		assert(not consumer.isExecuting())

	def testHasAborted(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(5, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)
		consumer.abort()
		assert(consumer.hasAborted())

	def testPopAll(self):
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)

		self.insertTestTuple()
		self.insertTestTuple()

		while consumer.isExecuting():
			time.sleep(5)

		results = consumer.pop()
		assert(len(results) == 2)

	def testPopLimit(self):
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)

		self.insertTestTuple()
		self.insertTestTuple()

		while consumer.isExecuting():
			time.sleep(5)

		results = consumer.pop(1)
		assert(len(results) == 1)
		results = consumer.pop(1)
		assert(len(results) == 1)
		results = consumer.pop(1)
		assert(len(results) == 0)

	def testEndOfResults(self):
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)

		self.insertTestTuple()
		self.insertTestTuple()
		numResults = 0
		
		while 1:
			results = consumer.pop(1)
			numResults += len(results)
			if results.isEndOfResults():
				break;			
			time.sleep(5)

		assert numResults == 2
		
	def testCount(self):
		"""
		Check that count returns the right number of tuples. Note that we
		are hoping here that nothing gets inserted between count() and pop()
		which shouldn't happen, but could in theory and would cause the test
		to fail.
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)

		time.sleep(10)
		self.insertTestTuple()
		time.sleep(1)
		self.insertTestTuple()
		time.sleep(5)

		while consumer.isExecuting():
			num_results = consumer.count()
			results = consumer.pop()
			assert(num_results == len(results))
			time.sleep(5)

        def testClose(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.close()
		caught = 0
		try:
			consumer.start()
		except:
			caught = 1
		assert caught

        def testDestroy(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.destroy()
		caught = 0
		try:
			consumer.start()
		except:
			caught = 1
		assert caught
		
        def testgetEndpoint(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		re = consumer.getEndpoint()
		url = re.getUrl()
		id = re.getResourceId()

	def testGetTerminationInterval(self):
		ti = rgma.TimeInterval(25, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		i = consumer.getTerminationInterval()
		assert (ti.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

#	def testGetVersion(self):
#		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
#		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
#		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
#		version = consumer.getVersion()
#		splitVersion = version.split('.')
#		assert (len(splitVersion) == 3)
#		major = int(splitVersion[0])
#		assert(major >= 0)
#		minor = int(splitVersion[1])
#		assert(minor >= 0)
#		patch = int(splitVersion[2])
#		assert(patch >= 0)
		
	def testSetTerminationInterval(self):
		ti = rgma.TimeInterval(15, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		tiNew = rgma.TimeInterval(17, rgma.Units.MINUTES)
		consumer.setTerminationInterval(tiNew)
		i = consumer.getTerminationInterval()
		assert (tiNew.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

	def testShowSignOfLife(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.showSignOfLife()

class PrimaryProducerTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(PrimaryProducerTest,'test')

	def testCreateWithVOs(self):
		"""
		Tests a deprecated constructor param
		"""
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp, [])

	def testCreateContinuous(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)

	def testCreateLatest(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.PrimaryProducer(ti, pp)

	def testCreateHistory(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.HISTORY)
		p = rgma.PrimaryProducer(ti, pp)
		
	def testCreateContinuousDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)

	def testCreateLatestDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.PrimaryProducer(ti, pp)

	def testCreateHistoryDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS + rgma.QueryType.HISTORY)
		p = rgma.PrimaryProducer(ti, pp)		

	def testDeclareTable(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)

	def testGetLatestRetentionPeriod(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.LATEST)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		lrp = p.getLatestRetentionPeriod(self.TEST_TABLE)
		assert (rp.getValueAs(rgma.Units.SECONDS) == lrp.getValueAs(rgma.Units.SECONDS))

	def testGetHistoryRetentionPeriod(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.HISTORY)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		hrp = p.getHistoryRetentionPeriod(self.TEST_TABLE)
		assert (rp.getValueAs(rgma.Units.SECONDS) == hrp.getValueAs(rgma.Units.SECONDS))

	def testInsert(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))

	def testInsertList(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		i1 = "INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string1', 1.2, 3)" % \
			 (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId)
		i2 = "INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string2', 4.5, 6)" % \
			 (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId)
		i3 = "INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string3', 7.8, 9)" % \
			 (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId)
		p.insertList([i1, i2, i3])

        def testClose(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		p.close()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti, ti)
		except:
			caught = 1
		assert caught

        def testDestroy(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		p.destroy()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti, ti)
		except:
			caught = 1
		assert caught
		
        def testgetEndpoint(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		re = p.getEndpoint()
		url = re.getUrl()
		id = re.getResourceId()

	def testGetTerminationInterval(self):
		ti = rgma.TimeInterval(25, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		i = p.getTerminationInterval()
		assert (ti.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

#	def testGetVersion(self):
#		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
#		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
#		p = rgma.PrimaryProducer(ti, pp)
#		version = p.getVersion()
#		splitVersion = version.split('.')
#		assert (len(splitVersion) == 3)
#		major = int(splitVersion[0])
#		assert(major >= 0)
#		minor = int(splitVersion[1])
#		assert(minor >= 0)
#		patch = int(splitVersion[2])
#		assert(patch >= 0)

	def testSetTerminationInterval(self):
		ti = rgma.TimeInterval(15, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		tiNew = rgma.TimeInterval(17, rgma.Units.MINUTES)
		p.setTerminationInterval(tiNew)
		i = p.getTerminationInterval()
		assert (tiNew.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

	def testShowSignOfLife(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		p.showSignOfLife()
		
class OnDemandProducerTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(OnDemandProducerTest,'test')

	def testCreateHttp(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)

	def testCreateHttps(self):
		uri = 'https://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)

	def testCreateSocket(self):
		uri = 'socket://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)

	def testDeclareTable(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		p.declareStaticTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId))

        def testClose(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		p.close()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti, ti)
		except:
			caught = 1
		assert caught

        def testDestroy(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		p.destroy()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti, ti)
		except:
			caught = 1
		assert caught
		
        def testgetEndpoint(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		re = p.getEndpoint()
		url = re.getUrl()
		id = re.getResourceId()

	def testGetTerminationInterval(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(25, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		i = p.getTerminationInterval()
		assert (ti.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

#	def testGetVersion(self):
#		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
#		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
#		p = rgma.PrimaryProducer(ti, pp)
#		version = p.getVersion()
#		splitVersion = version.split('.')
#		assert (len(splitVersion) == 3)
#		major = int(splitVersion[0])
#		assert(major >= 0)
#		minor = int(splitVersion[1])
#		assert(minor >= 0)
#		patch = int(splitVersion[2])
#		assert(patch >= 0)

	def testSetTerminationInterval(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(15, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		tiNew = rgma.TimeInterval(17, rgma.Units.MINUTES)
		p.setTerminationInterval(tiNew)
		i = p.getTerminationInterval()
		assert (tiNew.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

	def testShowSignOfLife(self):
		uri = 'http://example.com:1234/my/ondemand/producer/'
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p = rgma.OnDemandProducer(ti, uri)
		p.showSignOfLife()
		
class SecondaryProducerTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(SecondaryProducerTest,'test')

	def testCreateContinuous(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)

	def testCreateLatest(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.SecondaryProducer(ti, pp)

	def testCreateHistory(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.HISTORY)
		p = rgma.SecondaryProducer(ti, pp)		

	def testCreateContinuousDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)

	def testCreateLatestDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.SecondaryProducer(ti, pp)

	def testCreateHistoryDatabase(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS + rgma.QueryType.HISTORY)
		p = rgma.SecondaryProducer(ti, pp)		

	def testDeclareTable(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.SecondaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp)

	def testGetHistoryRetentionPeriod(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.HISTORY)
		p = rgma.SecondaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp)
		hrp = p.getHistoryRetentionPeriod(self.TEST_TABLE)
		assert (rp.getValueAs(rgma.Units.SECONDS) == hrp.getValueAs(rgma.Units.SECONDS))
		
        def testClose(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		p.close()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti)
		except:
			caught = 1
		assert caught

        def testDestroy(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		p.destroy()
		caught = 0
		try:
			p.declareTable(self.TEST_TABLE, "", ti)
		except:
			caught = 1
		assert caught
		
        def testgetEndpoint(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		re = p.getEndpoint()
		url = re.getUrl()
		id = re.getResourceId()

	def testGetTerminationInterval(self):
		ti = rgma.TimeInterval(25, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		i = p.getTerminationInterval()
		assert (ti.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

#	def testGetVersion(self):
#		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
#		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
#		p = rgma.SecondaryProducer(ti, pp)
#		version = p.getVersion()
#		splitVersion = version.split('.')
#		assert (len(splitVersion) == 3)
#		major = int(splitVersion[0])
#		assert(major >= 0)
#		minor = int(splitVersion[1])
#		assert(minor >= 0)
#		patch = int(splitVersion[2])
#		assert(patch >= 0)

	def testSetTerminationInterval(self):
		ti = rgma.TimeInterval(15, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		tiNew = rgma.TimeInterval(17, rgma.Units.MINUTES)
		p.setTerminationInterval(tiNew)
		i = p.getTerminationInterval()
		assert (tiNew.getValueAs(rgma.Units.SECONDS) == i.getValueAs(rgma.Units.SECONDS))

	def testShowSignOfLife(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.SecondaryProducer(ti, pp)
		p.showSignOfLife()

class SchemaTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(SchemaTest,'test')

	def testGetAllTables(self):
		s = rgma.Schema()
		tables = s.getAllTables()
		found = 0
		for table in tables:
			if table == self.TEST_TABLE:
				found = 1

		assert found

	def testGetTableDefinition(self):
		s = rgma.Schema()
		table = s.getTableDefinition(self.TEST_TABLE)
		assert table.name == self.TEST_TABLE
		assert len(table.columns) == 6
		for column in table.columns:
			if column.name == self.TEST_ID_COLUMN:
				assert column.type == rgma.Types.VARCHAR
				assert column.primaryKey
				assert column.notNull
			elif column.name == self.TEST_STRING_COLUMN:
				assert column.type == rgma.Types.VARCHAR
				assert not column.primaryKey
			elif column.name == self.TEST_REAL_COLUMN:
				assert column.type == rgma.Types.REAL
				assert not column.primaryKey
			elif column.name == self.TEST_INT_COLUMN:
				assert column.type == rgma.Types.INTEGER
				assert not column.primaryKey
			elif column.name == self.DATE_COLUMN:
				assert column.type == rgma.Types.DATE
				assert not column.primaryKey
				assert column.notNull
			elif column.name == self.TIME_COLUMN:
				assert column.type == rgma.Types.TIME
				assert not column.primaryKey
				assert column.notNull
			else:
				assert False

	def testCreateDropTable(self):
		s = rgma.Schema()
		tables = s.getAllTables()
		i = 0
		while 1:
			tableName = self.TEST_TABLE + str(i)
			if tableName not in tables: break
			i += 1

		s.createTable("CREATE TABLE %s (%s VARCHAR(255) PRIMARY KEY, %s VARCHAR(255), %s REAL, %s INTEGER)" %
			      (tableName, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN), [])
		tables = s.getAllTables()
		assert tableName in tables
		table = s.getTableDefinition(tableName)
		assert table.name == tableName
		assert len(table.columns) == 6
		for column in table.columns:
			if column.name == self.TEST_ID_COLUMN:
				assert column.type == rgma.Types.VARCHAR
				assert column.primaryKey
				assert column.notNull
			elif column.name == self.TEST_STRING_COLUMN:
				assert column.type == rgma.Types.VARCHAR
				assert not column.primaryKey
			elif column.name == self.TEST_REAL_COLUMN:
				assert column.type == rgma.Types.REAL
				assert not column.primaryKey
			elif column.name == self.TEST_INT_COLUMN:
				assert column.type == rgma.Types.INTEGER
				assert not column.primaryKey
			elif column.name == self.DATE_COLUMN:
				assert column.type == rgma.Types.DATE
				assert not column.primaryKey
				assert column.notNull
			elif column.name == self.TIME_COLUMN:
				assert column.type == rgma.Types.TIME
				assert not column.primaryKey
				assert column.notNull
			else:
				assert False
		s.dropTable(tableName)
		tables = s.getAllTables()
		assert tableName not in tables

		
class RegistryTest(unittest.TestCase, TestBase):	

	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(RegistryTest,'test')

	def testGetAllProducersForTable(self):
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		re = p.getEndpoint()

		time.sleep(2)
		
		ra = rgma.Registry()
		producers = ra.getAllProducersForTable(self.TEST_TABLE)
		found = 0
		for producer in producers:
			e = producer.getEndpoint()
			if (e.getUrl() == re.getUrl()) and \
			   (e.getResourceId() == re.getResourceId()):
				found = 1

		assert found
		
class ReconnectTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(ReconnectTest,'test')

 	def testReconnectSecondaryProducer(self):
 		"""
 		Create a secondary producer, delete the object
 		then reconnect to it and declare a table
 		"""
 		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
 		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
 		p = rgma.SecondaryProducer(ti, pp)
 		re = p.getEndpoint()

		p = None
		time.sleep(2)

 		p = rgma.reconnect(rgma.SecondaryProducer, re)
 		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
 		p.declareTable(self.TEST_TABLE, '', rp)

 	def testReconnectPrimaryProducer(self):
 		"""
 		Create a primary producer, delete the object
 		then reconnect to it and declare a table
 		"""
 		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
 		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
 		p = rgma.PrimaryProducer(ti, pp)
 		re = p.getEndpoint()

		p = None
		time.sleep(2)

 		p = rgma.reconnect(rgma.PrimaryProducer, re)
 		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
 		p.declareTable(self.TEST_TABLE, '', rp, rp)
		
 	def testReconnectConsumer(self):
 		"""
 		Create a consumer, insert a tuple for it to receive. The delete
		the object, reconnect to it and pop the tuple
 		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		re = consumer.getEndpoint()
		
		consumer.start(timeout)
		time.sleep(10)
		self.insertTestTuple()
		time.sleep(5)
			
		consumer = None
		time.sleep(2)

 		consumer = rgma.reconnect(rgma.Consumer, re)
		num_results = 0
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)
			
		assert num_results == 1

class ProducerConsumerTest(unittest.TestCase, TestBase):
	
	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(ProducerConsumerTest,'test')
		
	def testContinuousQuery(self):
		"""
		Check that a continuous consumer gets a single tuple inserted while
		it is running
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer.start(timeout)

		time.sleep(10)
		self.insertTestTuple()
		time.sleep(5)

		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)
			
		results = consumer.pop()
		num_results += len(results)

		assert(num_results == 1)

	def testContinuousHistoryQuery(self):
		"""
		Check that a continuous consumer gets a single tuple inserted before it
		was started, but only if a continuous+history query was specified
		"""
		self.insertTestTuple()
		time.sleep(5)
		
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)

		consumer1 = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)
		consumer2 = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS, rgma.TimeInterval(5, rgma.Units.MINUTES)), query)

		consumer1.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer1.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)

		assert(num_results == 0)

		consumer2.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer2.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)

		assert(num_results == 1)

	def testRetentionPeriodContinuous(self):
		"""
		Check that tuples are not available from a continuous primary producer after their
		history retention period
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.SECONDS)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))

		time.sleep(10)
		
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS, rgma.TimeInterval(5, rgma.Units.MINUTES)), query)

		consumer.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)
			
		assert(num_results == 0)

	def testDirectedQuery(self):
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.CONTINUOUS), query)

		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
 		rp = rgma.TimeInterval(1, rgma.Units.MINUTES)
		p1 = rgma.PrimaryProducer(ti, pp)
 		p1.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p2 = rgma.PrimaryProducer(ti, pp)
 		p2.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		
		re = p1.getEndpoint()
		consumer.start(timeout, [re,])

		time.sleep(5)
		p1.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		p2.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string2', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		time.sleep(5)

		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)

		# check we only get results from producer 1
		assert(num_results == 1)
		

	def testLatestPrimaryProducer(self):
		"""
		Check that a latest primary producer works
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.CONTINUOUS + rgma.QueryType.LATEST)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		time.sleep(1)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string2', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))

		time.sleep(10)
		
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.LATEST), query)

		consumer.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if len(results) > 0:
				tuple = results[len(results) - 1]
			if results.isEndOfResults(): break
			time.sleep(5)
			
		# Check we get only the latest tuple
		assert(num_results == 1)
		assert(tuple[1] == 'string2')

	def testLatestSecondaryProducer(self):
		"""
		Check that a latest secondary producer picks up tuples
		inserted into a continuous primary producer
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.LATEST)
		sp = rgma.SecondaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		sp.declareTable(self.TEST_TABLE, '', rp)

		time.sleep(5)
		
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string2', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))

		time.sleep(30)
		
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(40, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.LATEST), query)

		consumer.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if len(results) > 0:
				tuple = results[len(results) - 1]
			if results.isEndOfResults(): break
			time.sleep(5)
			
		# Check we get only the latest tuple
		assert(num_results == 1)
		assert(tuple[1] == 'string2')

	def testHistorySecondaryProducer(self):
		"""
		Check that a latest secondary producer picks up tuples
		inserted into a continuous primary producer
		"""
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), rgma.QueryType.HISTORY)
		sp = rgma.SecondaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		sp.declareTable(self.TEST_TABLE, '', rp)

		time.sleep(5)
		pp = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.MEMORY), rgma.QueryType.CONTINUOUS)
		p = rgma.PrimaryProducer(ti, pp)
		rp = rgma.TimeInterval(5, rgma.Units.MINUTES)
		p.declareTable(self.TEST_TABLE, "WHERE %s='%s'" % (self.TEST_ID_COLUMN, self.uniqueId), rp, rp)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))
		time.sleep(1)
		p.insert("INSERT INTO %s (%s, %s, %s, %s) values ('%s', 'string2', 1.2, 3)" \
			 % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.TEST_STRING_COLUMN, self.TEST_REAL_COLUMN, self.TEST_INT_COLUMN, self.uniqueId))

		time.sleep(10)
		
		ti = rgma.TimeInterval(2, rgma.Units.MINUTES)
		timeout = rgma.TimeInterval(30, rgma.Units.SECONDS)
		query = "SELECT * FROM %s WHERE %s='%s'" % (self.TEST_TABLE, self.TEST_ID_COLUMN, self.uniqueId)
		consumer = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.HISTORY), query)

		consumer.start(timeout)
		num_results = 0;
		
		while 1:
			results = consumer.pop(10)
			num_results += len(results)
			if results.isEndOfResults(): break
			time.sleep(5)

		# Check we get both tuples
		assert(num_results == 2)

class ExceptionsTest:

	def setUp(self):
		TestBase.__init__(self)
	
	def suite(self):
		return unittest.makeSuite(ExceptionsTest,'test')

	def testUnknownResourceException(self):
		url = rgma.Properties().getProperty("Consumer")[0]
		re = rgma.ResourceEndpoint(url, "0")
		c = rgma.reconnect(rgma.Consumer, re)
		caught = False
		try:
			c.showSignOfLife()
		except rgma.UnknownResourceException:
			caught = True

		assert caught

	def testRemoteException(self):
		url = "http://r-gma.org/invalid/url"
		re = rgma.ResourceEndpoint(url, "0")
		c = rgma.reconnect(rgma.Consumer, re)
		caught = False
		try:
			c.showSignOfLife()
		except rgma.RemoteException:
			caught = True

		assert caught

	def testRGMAException(self):
		caught = False
		ti = rgma.TimeInterval(1, rgma.Units.MINUTES)
		try:
			c = rgma.Consumer(ti, rgma.QueryProperties(rgma.QueryType.LATEST), "SELECT * from noSuchTable")
		except rgma.RGMAException:
			caught = True

		assert caught
		
if __name__ == "__main__":
	unittest.main()
